% 
%     [ Demod1, Demod2, PRG, Es ] = DemodRF( ifo, fldIDs, RF, modIndex, error )
% 
%     how to use
%         ifo = IFOdefinition;
%         ifo.lock;
%         [Demod1, Demod2, PRG, Es] = demod( ifo, CRlock, {'PRM-AR-o', dec1; 'BS-ARy-o', dec2}, 9e6, 0.1, 1e-4);
%         dec = 0 for circular, dec = 1 for quadrand
%         dec = function handle val = dec(CR, SB); val can be multiple values like quandrant
% 
%     what is calculated
%         this calculates the error signal demodulated with RF and 2RF
%         SB is added as Etot = ECR * exp( i G Cos[2 pi RF t] )
% 
%     input
%         fldIDs : list of fields to calculate the error signals
%         RF : modulation and demoulation frequency
%         modIndex : modulation index as defined above [G=1]
%         errorCR, errorBS : accuracy [def=1e-6,1e-4], if errorCR < 0, CR calculation has been done, and don't calculate
% 
%     output
%         Idemo(1:Number of fields) = real( Demod1/2( fieldID ) )
%         Qdemo(1:Number of fields) = imag( Demod1/2( fieldID ) )
%         Demod1/2 could be an array depending on the detector
%         For detector 1, quadrand data (x>0,y>0), (x>0,y<0), (x<0,y<0), (x<0,y>0) are returned in this order
%         When function_handle is used, any number of values could be returned.
%              AIS = Integal(x,y) | Ecr + Eusb exp(i gamma RF t) + Elsb exp(-i gamma RF t) |^2
%                  = CDemod1 cos(Omega t + phi1) + CDemod2 cos(2 Omega t + phi2)
%              DemodSig = Integral(t) AIS(t) Cos(Omega t - demod1)
%         PRG(1:3) = PRG of CR, USB and LSB
%         Es(1:number of field IDs) = struct ( CR, USB, LSB, x, y )
%
%     May 24, 2018
%     This is the backend of demoRF in FFTIFO.
%     DemodRFQuad is an example of adding new shapes of detectors.
%
%     May 21, 2018
%     added second argument in the field list specifying the detector shape
%
%     April 3, 2018
%     Support 2 Omega demod
% 
%     February 27, 2018
%     Integration of Demod.m in FFTIFO
% 
%     May 23, 2017
%     2 Omega demodulation added
%     moduleation index included
% 
%     January 12, 2017  Hiro YamamotoASCB(ic,1:4) = fp.DemodRF( {'ITM-AR-o',@FPIFO0.DemodRFQuad}, RF, 1, error );
%     First release
%

function [ Demod1, Demod2, PRG, Es ] = DemodRFFunc( ifo, fldIDs, RF, modIndex, errorCR, errorSB )

RFsave = ifo.getRF;

BSexist = ifo.optics.findNamedOptic('BS') > 0;

% fldIDs = 'ITM-AR-o' or 12 or {'ITM-AR-o'; 'BS-ARy-o'} or {ID1, det1; ID2, det2; ...}
if ischar( fldIDs ) || isnumeric( fldIDs )  % 'ITM-AR-o' or 12
    fldIDs = {fldIDs};
end
nFlds = size(fldIDs,1);
fIDs = zeros(nFlds,1);
for id = 1:nFlds
    fIDs(id) = ifo.fields.findNamedField( fldIDs{id} );
end
PRG = zeros(1,3);

fprintf('\n\n--> CR for demod\n');
ifo.restoreFields(1); % if no saved data, just don't do any
% if error < 0, CR calculation has been done, so don't redo
if ifo.getRF ~= 0 || errorCR > 0
    ifo.setRF(0);
    ifo.calc('errorLimit', abs(errorCR), 'Nprint', 20);
end
if BSexist
    PRG(1) = ifo.power('BS-HRx-o');
end
ifo.saveFields(1);
[E, x, y] = ifo.getField(fIDs(1));
Es = repmat( struct('CR', E, 'USB', E, 'LSB', E, 'x', x, 'y', y ), nFlds );
for id = 2 : nFlds
    [Es(id).CR, Es(id).x, Es(id).y] = ifo.getField(fIDs(id));
end

fprintf('\n----> USB for demod\n');
ifo.setRF(RF);
ifo.restoreFields(2);  % if no saved data, just don't do any
ifo.calc('errorLimit', errorSB, 'Nprint', 20);
for id = 1 : nFlds
    Es(id).USB = 0.5i*ifo.getField(fIDs(id))*modIndex;
end
if BSexist
    PRG(2) = ifo.power('BS-HRx-o');
end
ifo.saveFields(2);

fprintf('\n----> LSB for demod\n');
ifo.setRF(-RF);
ifo.restoreFields(3);  % if no saved data, just don't do any
ifo.calc('errorLimit', errorSB, 'Nprint', 20);
for id = 1 : nFlds
    Es(id).LSB = 0.5i*ifo.getField(fIDs(id))*modIndex;
end
if BSexist
    PRG(3) = ifo.power('BS-HRx-o');
end
ifo.saveFields(3);

ds = (x(2)-x(1))^2;
if size(fldIDs,2) > 1
    [xg, yg] = meshgrid( x, y );
end
Demod1 = zeros(nFlds,1);
Demod2 = zeros(nFlds,1);
for id = 1:nFlds
    circular = true;
    if size(fldIDs,2) > 1
        if isa( fldIDs{id,2}, 'function_handle' )
            [val1, val2] = fldIDs{id,2}( Es(id), xg, yg );
            Demod1(id,1:length(val1)) = val1;
            Demod2(id,1:length(val2)) = val2;
            circular = false;
         elseif fldIDs{id,2} == 1  % quadrant
             [val1, val2] = DemodRFQuad( Es(id), xg, yg );
             Demod1(id,1:length(val1)) = val1;
             Demod2(id,1:length(val2)) = val2;
             circular = false;
        end
    end
    if circular
        Demod1(id,1) = 2 * sum(sum(conj( Es(id).CR ) .* Es(id).LSB + Es(id).CR .* conj(Es(id).USB ))) * ds;
        Demod2(id,1) = 2 * sum(sum(Es(id).LSB .* conj(Es(id).USB))) * ds;
    end
end

ifo.setRF(RFsave);
end

% use this as an exmple of defining external demod function
function [Demod1, Demod2] = DemodRFQuad( Es, xg, yg )
ds = (xg(1,2)-xg(1,1))*(yg(2,1)-yg(1,1));
for isec = 1 : 4
    switch isec
        case 1
            mask = (xg >= 0) & (yg >= 0);
        case 2
            mask = (xg >= 0) & (yg < 0);
        case 3
            mask = (xg < 0) & (yg < 0);
        otherwise
            mask = (xg < 0) & (yg >= 0);
    end
    Demod1(isec) = 2 * sum(sum( (conj( Es.CR ) .* Es.LSB + Es.CR .* conj(Es.USB) ).* mask ) ) * ds;
    Demod2(isec) = 2 * sum(sum( (Es.LSB .* conj(Es.USB)) .* mask) ) * ds;
end
end