%
% ModalModelTK : toolkit to calculate modal model related functions
%
% November 4, 2019
%    many functions which have k values as the last agument is modifed to (,,,varargin)
%    so that the test ~exist('k', 'var') is done only at the last function which actually use k
%
% Sep. 24, 2019
%    function [wL, RinvL] = RinvWZ2WRinv( Rinv, w, L, k ) : same as RinvWZ2RinvW, just wL first
%             [RinvL, wL] = RinvWZ2RinvW( Rinv, w, L, k )
%
% April 3, 2019
%   function [z, z0] = RinvW2zz0( Rinv, w, k )
%   function [RinvZ, wZ] = RinvWZ2RinvW( Rinv, w, L, k ) <= changed z to L to clarify the meaning of z
%   function qinv = RinvW2Qinv( Rinv, w, k )
%   These functions use k, instead of lambda.
%
% March 6, 2019
%    function [w1, w2] = w1FP( RoC1inv, RoC2inv, Lcav, k )
%       now retursn w1 and w2 together
%
% April 13, 2018
%  function L = adjustL4RoC( invRoC, M2, M1, qInv ) added
%     qInvOut = M2 MProp(L) M1 qInvIn
%     L is adjusted to make real(qInvOut) = invRoC
%
% April 8, 2018
%    function [q1, q2] = FPparamQ( RoC1inv, RoC2inv, Lcav ) returns q1 and q2
%
% September 27, 2016
%    reflAR_ABCD is replaced by reflN_ABCD.
%    This is to be used for the reflection in a substrate with refractive index n.
%    This is to be used for the reflection of the field from the AR side.
%    Mirror4x4 and Mirror2x2 used reflAR, repalced by reflN with proper sign change
%
% September 27, 2016
%    default k or lam defined as constant properties, so k does not need to be provided
%
% April 19, 2016
%    two funnctions added
%        function [y1, y2, theta] = cavityAxis( RoC1Inv, RoC2Inv, Lcav, theta1, theta2 )
%        function theta0 = divAngle( RoC1Inv, RoC2Inv, Lcav, lam )
%
% sometime at the beginning of 2016
%   wave length or k is passed when needed, so no default k or lambda in MMTK
%
% September 2, 2015
%   name convension is to use inv when in or out variables are inv
%
% September 1st, 2015
%   lam0 and kLam0 are defined in this class, not in ConstantTK.
%   For now, there is no default argument
%
% August 30, 2015
%   RinvW2zz0 and RinvWZ2RinvW to calulate quantities using R and w (and k)
%
% February 24, 2012
%   Consisten use of inv, all names are Xinv, like Rinv, qinv, etc
%   Remove redundant expressions
%
% August 5, 2010
%   Use Rinv (=1/R) instead of R as arguments.
%   f number is removed, except RinvBack.
%
% June 24, 2010  Hiro Yamamoto  hiro@ligo.caltech.edu
%   first release
%

classdef MMTK
    
    properties (Constant)
        lam0 = 1.064e-6; % default CR laser wave length
        kLam0 = 2*pi / MMTK.lam0;
    end
    
    methods (Static)
        
        %% MM bacis
        % effective curvature seen from back side of a mirror
        % with refractive index n and internal f number invF
        % Rm > 0 for concave mirror
        % Use this for the second argument of FPparam to
        % calculate parameters of a Michelson cavity
        function Rinv = RinvBack( Rinv, n, Finv )
            Rinv = - n * Rinv + Finv;
        end
        
        % modal paramters in a FP cavity defined by curvatures and the cavity length.
        % returns Rayleigh range and the front mirror z
        function [z, z0] = FPparam( RoC1inv, RoC2inv, Lcav )
            g1 = 1 - Lcav * RoC1inv;
            g2 = 1 - Lcav * RoC2inv;
            ggg = g1 + g2 - 2*g1.*g2;
            z = Lcav * g2.*(1 - g1)./ggg;
            z0 = Lcav * sqrt( g1.*g2.*(1 - g1.*g2) ) ./ abs( ggg );
        end
        
        % q of a FP cavity
        % these qs going toward the mirror
        function [q1, q2] = FPparamQ( RoC1inv, RoC2inv, Lcav )
            [z, z0] = MMTK.FPparam( RoC1inv, RoC2inv, Lcav );
            q1 = z + 1i * z0;
            q2 = -(z - Lcav) + 1i * z0;
        end
        
        % beam size on front mirror
        function [w1, w2] = w1FP( RoC1inv, RoC2inv, Lcav, k )
            if ~exist( 'k', 'var' )
                k = MMTK.kLam0;
            end
            [z, z0] = MMTK.FPparam( RoC1inv, RoC2inv, Lcav );
            w0 = sqrt(2*z0/k);
            w1 = w0*sqrt( 1 + (z/z0)^2 );
            if nargout == 2
                w2 = w0*sqrt( 1 + ((Lcav-z)/z0)^2 );
            end
        end
        
        function RoC = wL2RoC( w, L, gPlusMin, k )
            % RoC = wL2RoC( w, L, gPlusMin, k )
            % RoC of a symmetric cavity which makes the beam size to be w,
            % gPlusMin is the choice of positive or negative g
            if ~exist( 'k', 'var' )
                k = MMTK.kLam0;
            end
            %             if ~exist( 'plusMinus', 'var' )
            %                 gPlusMin = -1;
            %             end
            A = 4 * L / (k^2 * w^4);
            RoC = sqrt( 1 - A * L );
            if gPlusMin > 0
                RoC = ( 1 + RoC ) / A;
            else
                RoC = ( 1 - RoC ) / A;
            end
        end
        
        function eta = gouy00FP( RoC1inv, RoC2inv, Lcav )
            % eta = gouy00FP( RoC1inv, RoC2inv, Lcav )
            % gouy phase of a FP cavity
            [z, z0] = MMTK.FPparam( RoC1inv, RoC2inv, Lcav );
            eta = atan2( z, z0 ) - atan2( z - Lcav, z0 );
        end
        
        % gouy phase of a field from q = z + i z0 to q + L
        function eta = gouyL( q, L )
            eta = atan2( real(q)+L, imag(q) ) - atan2( real(q), imag(q) );
        end
        
        % function [y1, y2, theta] = cavityAxis( RoC1Inv, RoC2Inv, Lcav, theta1, theta2 )
        % resonating beam offset on mirror 1 and mirror 2 when mirrors are tilted
        % theta1 and theta2 are pitch angle around x axis and the offset is along y axis
        % where +y is always pointing upward (same for both mirrors), but
        % +x direction is defined by +z (beam goint out) and +y and are oppotite for mirror 1 and mirror 2
        % theta is the angle of the beam going from mirror 1 toward mirror 2
        function [y1, y2, theta] = cavityAxis( RoC1Inv, RoC2Inv, Lcav, theta1, theta2 )
            y1 = ( Lcav * RoC2Inv*theta1 - theta1 - theta2 ) / (RoC1Inv + RoC2Inv - Lcav*RoC1Inv*RoC2Inv);
            y2 = ( Lcav * RoC1Inv*theta2 - theta1 - theta2 ) / (RoC1Inv + RoC2Inv - Lcav*RoC1Inv*RoC2Inv);
            theta = (y2 - y1) / Lcav;
        end
        
        % function theta0 = divAngle( RoC1Inv, RoC2Inv, Lcav, lam )
        % divergence angle on mirror 1, lam / pi w, in a FP cavity
        function theta0 = divAngle( RoC1Inv, RoC2Inv, Lcav, lam )
            theta0 = lam / pi / MMTK.w1FP( RoC1Inv, RoC2Inv, Lcav, 2*pi/lam );
        end
        
        %% ABCD matrix toolkit
        
        % ABCD operation on q
        function qp = ABCD_q( M, q )
            qp = ( M(1,1) * q + M(1,2) ) ./ ( M(2,1) * q + M(2,2) );
        end
        
        % ABCD operation on 1/q
        function qpinv = ABCD_qinv( M, qinv )
            qpinv = ( M(2,1) + M(2,2)*qinv ) ./ ( M(1,1) + M(1,2)*qinv );
        end
        
        % when qInv' is given for qInv' = ABCD(qInv), qInv is calculated
        function qinv = ABCD_qinvpRev( M, qinvp )
            qinv = ( M(2,1) - M(1,1)*qinvp ) ./ ( M(1,2)*qinvp - M(2,2) );
        end
        
        % ABCD operation on (z, z0)
        function [zp, z0p] = ABCD_zz0( M, z, z0 )
            qp = ABCD_q( M, z + 1i*z0 );
            zp = real( qp );
            z0p = imag( qp );
        end
        
        % ABCD operation on (1/R and w)
        % varargin = k if speficied
        function [Rpinv, wp] = ABCD_rw( M, Rinv, w, varargin )
            [Rpinv, wp] = MMTK.qinv2RinvW( MMTK.ABCD_qinv( M, Rinv - 1i*2/k/w^2 ), varargin );
        end
        
        % Siegman Table 15.1 (a)
        % propagation through a substrate with refractive index n
        function M = Prop_ABCD( L, n )
            if ~exist('n', 'var')
                n = 1;
            end
            M = [ 1, L/n; 0, 1 ];
        end
        
        % Siegman Table 15.1 (b)
        % lens specified by f number.
        % f > 0 corresponds to converging lens
        % Lens(finv1) * Lens(finv2) = Lens(finv1+finv2), i.e., commutable
        function M = LensF_ABCD( finv )
            M = [ 1, 0; -finv, 1 ];
        end
        
        % Siegman Table 15.1 (b) f specified by R and n
        % lens through a substrate with refractive index n and
        % one side RoCinv ( positive for concave )
        function M = LensR_ABCD( Rinv, n )
            M = MMTK.LensF_ABCD( -(n-1)*Rinv );
        end
        
        % Siegman Table 15.1 (d)
        % reflection with angle of incidence
        % plane = 0 : tangential (in the plane of incidence)
        %       = 1 : sagittal (perpendicular to plane of incidence)
        function M = Refl_ABCD( Rinv, theta, plane )
            if ~exist('plane', 'var')
                plane = 0;
            end
            if ~exist('theta', 'var')
                theta = 0;
            end
            if plane == 0
                cosCor = cos( theta );
            else
                cosCor = 1/cos( theta );
            end
            M = [ 1, 0; -2*Rinv/cosCor, 1];
        end
        
        % Siegman Table 15.1 (c) with refractive index
        % Reflection in the side with finite refractive index
        % Rinv > 0 for concave seen from the incident direction
        function M = ReflN_ABCD( Rinv, n )
            if ~exist( 'n', 'var' )
                n = 1;
            end
            M = [ 1, 0; -2*n*Rinv, 1];
        end
        
        % Siegman Table 15.1 (e)
        % Curved dielectric interface, normal incidence
        % R > 0 for concave surface
        function M = CurvedInterface_ABCD( Rinv, n1, n2 )
            M = [ 1, 0; (n2-n1)*Rinv, 1];
        end
        
        % reflection from AR side, substrate refractive index of n
        % R are positive for concave
        % go into a AR surface with curvature RinvAR
        % go through a substrate of thickness d
        % reflected by the HR side with curvature RinvHR
        % go through substrate and AR surface
        % M ~= [1, 2d/n; 2(n-1) RinvAR + 2n RinvHR - 2/f, 1] + d invRoC * ...
        function M = ReflAR_ABCD( finv, RinvHR, RinvAR, n, d )
            M = MMTK.CurvedInterface_ABCD( -RinvAR, n, 1 ) * MMTK.Prop_ABCD( d, n ) * ...
                MMTK.LensF_ABCD( finv ) * MMTK.ReflN_ABCD( -RinvHR, n ) * MMTK.LensF_ABCD( finv ) * ...
                MMTK.Prop_ABCD( d, n ) * MMTK.CurvedInterface_ABCD( RinvAR, 1, n );
        end
        
        % eigen value of a cavity defined by M
        % q = (A q + B) / (C q + D) => q = ( (A-D)+-sqrt((D-A)^2+4 B C) ) / 2 C
        % sqrt() should be imaginary...
        function qEigen = EigenMode_ABCD( M )
            z = (M(1,1) - M(2,2)) / (2*M(2,1));
            z0Sq = -(M(1,1)-M(2,2))^2 - 4*M(1,2)*M(2,1);
            if z0Sq > 0
                z0 = sqrt( z0Sq ) / abs( 2*M(2,1) );
            else
                z0 = NaN;
            end
            qEigen = z + 1i * z0;
        end
        
        % constucting M by reflection and transmission
        % if number of argument is odd, the arguments will be for one way
        % even case : r1 p12 r2 p23 r3 p32 r2 p21
        % odd case  : r1 p12 r2 p23 r3 [p23] [r2] [p12]
        function M = ReflProp_ABCD( varargin )
            if mod(nargin,2) == 0
                M = MMTK.Refl_ABCD( varargin{1} ) * MMTK.Prop_ABCD( varargin{2} );
                if nargin > 2
                    M = M * MMTK.ReflProp_ABCD( varargin{3:end} );
                end
            else
                if nargin == 1
                    %                    varargin{1}
                    M = MMTK.Refl_ABCD( varargin{1} );
                else
                    M = MMTK.ReflProp_ABCD( varargin{1:end-1} ) * ...
                        MMTK.ReflProp_ABCD( varargin{end:-1:2} );
                end
            end
        end
        
        % qInvOut = M2 MProp(L) M1 qInvIn
        % L is adjusted to make real(qInvOut) = invRoC
        function L = adjustL4RoC( invRoC, M2, M1, qInv )
            qInv = MMTK.ABCD_qinv( M1, qInv );
            qIsq = real(qInv)^2 + imag(qInv)^2;
            AC = M2(1,1)*M2(2,1);
            ADBC = M2(1,1)*M2(2,2) + M2(1,2)*M2(2,1);
            AB = M2(1,1) * M2(1,2);
            BD = M2(1,2) * M2(2,2);
            qInvR = real(qInv);
            
            N2 = AC * qIsq;
            N1 = 2 * AC * qInvR + ADBC * qIsq;
            N0 = AC + ADBC * qInvR + BD * qIsq;
            D2 = M2(1,1)^2 * qIsq;
            D1 = 2*(M2(1,1)^2 * qInvR + AB * qIsq );
            D0 = M2(1,1)^2+ 2*AB*qInvR + M2(1,2)^2 * qIsq;
            
            C2 = N2 - D2*invRoC;
            C1 = N1 - D1*invRoC;
            C0 = N0 - D0*invRoC;
            
            B24AC = C1^2 - 4*C0*C2;
            L(1) = (-C1+sqrt(B24AC))/(2*C2);
            L(2) = (-C1-sqrt(B24AC))/(2*C2);
        end
        
        % adjust length, which is negative, to make the RoC(qOut) = RoC1
        % when qIn propagates through the cavity defined
        % inputs : RoC1 L12 RoC2 L23 ... L, qIn, dL
        % qOut = ABCD( RoC1, L12, ..., L ) * qIn
        function [len0, qOut] = AdjustModeMatch( varargin )
            adjust = find( [varargin{2:2:end-2}] < 0 ) * 2;
            L0 = -varargin{adjust};
            dl0 = varargin{nargin};
            dl = dl0;
            len0 = L0;
            lenMin = L0 - dl;
            lenMax = L0 + dl;
            invRoC = varargin{1};
            qIn = varargin{end-1};
            
            % remove the reflection by the mirror at qOut
            varargin{1} = 0;
            
            % look for length to make qOut to match with RoC1
            invRoCOut = zeros(1,2);
            dinvR = zeros(1,2);
            while (dl/L0) > 1e-10
                len(1) = len0 - dl;
                len(2) = len0 + dl;
                for iLen = 1:2
                    varargin{adjust} = len(iLen);
                    M = MMTK.ReflProp_ABCD( varargin{1:end-2} );
                    qOut = MMTK.ABCD_q( M, qIn );
                    invRoCOut(iLen) = real(1/qOut);
                    dinvR(iLen) = invRoCOut(iLen) - invRoC;
                end
                % at -dl => dLen(1), at +dl => dLen(2)
                dinvR12 = dinvR(1) - dinvR(2);
                % dR/R is small enough
                if abs( dinvR12 / invRoC ) < 1e-6
                    break;
                end
                % expected 0 crossing is len0 + off*dl
                dl = ( dinvR(1) + dinvR(2) ) / dinvR12 * dl;
                len0 = len0 + dl;
                dl = abs(dl);
                if len0 < lenMin
                    len0 = lenMin;
                    break;
                elseif len0 > lenMax
                    len0 = lenMax;
                    break;
                end
                dl = min( dl, dl0 );
            end
            if nargout > 1
                varargin{adjust} = len0;
                M = MMTK.ReflProp_ABCD( varargin{1:end-2} );
                qOut = MMTK.ABCD_q( M, qIn );
            end
        end
        
        %% conversion among various forms
        function [z, z0] = RinvW2zz0( Rinv, w, varargin )
            % [z, z0] = RinvW2zz0( Rinv, w, k )
            % calculete (z,z0) from Rinv and w
            [z, z0] = MMTK.q2zz0( 1/MMTK.RinvW2Qinv( Rinv, w, varargin ) );
        end
        
        function [wL, RinvL] = RinvWZ2WRinv( Rinv, w, L, varargin )
            % [wL, RinvL] = RinvWZ2W( Rinv, w, L, k )
            % this uses RinvWZ2RinvW, just return wL as the first argument
            [RinvL, wL] = MMTK.RinvWZ2RinvW( Rinv, w, L, varargin{:} );
        end
        
        function [RinvL, wL] = RinvWZ2RinvW( Rinv, w, L, varargin )
            % [RinvL, wL] = RinvWZ2RinvW( Rinv, w, L, k )
            % calculete (Rinv, w) at distane L from (Rinv, w)
            q = 1 / MMTK.RinvW2Qinv( Rinv, w, varargin{:} ) + L;
            [RinvL, wL] = MMTK.qinv2RinvW( 1./q, varargin{:} );
        end
        
        function qinv = RinvW2Qinv( Rinv, w, k )
            %	function qinv = RinvW2Qinv( Rinv, w, k )
            %   Rinv, w to qInv
            if ~exist( 'k', 'var' )
                k = MMTK.kLam0;
            end
            qinv = Rinv - 2i / (k * w^2);
        end
        
        % q to (z, z0)
        function [z, z0] = q2zz0( q )
            z = real(q);
            z0 = imag(q);
        end
        
        function [Rinv, w] = qinv2RinvW( qinv, varargin )
            %   function [Rinv, w] = qinv2RinvW( qinv, k )
            % varargin = k if specified
            % q to 1/R and w
            Rinv = real(qinv);
            if nargout == 2
                w = MMTK.qinv2W(qinv, varargin{:});
            end
        end
        
        % w from q
        function w = qinv2W( qinv, k )
            if ~exist( 'k', 'var' )
                k = MMTK.kLam0;
            end
            w = sqrt( -2./(k*imag(qinv)) );
        end
        
        % reverse direction ( q or qinv will work )
        % q : z + i z0 => -z + i z0 or
        % qinv : 1/R - i lambda/pi 1/w^2 => -1/R - i lambda/pi 1/w^2
        function qp = reverseDir( q )
            qp = -real( q ) + 1i*imag( q );
        end
    end
end
