%
% MirrorObj : basec lass of mirror object,
%             mirror2x2 (normal incident) and mirror4x4 (finite incident angle) are derived from this
%
%   invRoCAR should be set explicitly by hand like
%       mirID = obj.addMirror( ... )
%       obj.optics.optList(mirID).invRoCAR = 1/RoCofARofThisMirror
%
%   calculate reflection by a mirror, in and out are both internal format
%   inField is flipped left-right
%       outField = refl( obj, inPortIndex, inField )
%
%   calculate transmission through a mirror, in and out are both internal format
%   No field is flipped left-right
%       outField = trans( obj, inPortIndex, inField )
%
% April 10, 2020
%     print is modified to print optTheta and optPhi
%
% November 5th, 2019
%     tiltX and tiltY are added, they represents the direction of the vertical axis from the HR side.
%     This is used to support FP cavities with tilted cavity axis without worrying about the mode matching on ITM.
%     In refl (AR->AR) and in trans(HR->AR, AR->HR), beams are tilted by proper amount so that the field on the AR side 
%     is always along the original z axis, vertical to the AR surface, even when the FP cavity axis is tilted.
%
% To use this, add tiltOn calls as follows.
%        [thetaI,phiI,thetaE,phiE] = cavityManager.beam2tilt( 0, 0, dxE, dyE, invRoCIM, invRoCEM, Lcav );
%        obj.addMirror( 'IM', ..., 'optTheta', thetaI, 'optPhi', phiI);
%        obj.addMirror( 'EM', ..., 'optTheta', thetaE, 'optPhi', phiE);
%        obj.tiltOn( 'IM', -dxE/Lcav, dyE/Lcav );
%
% November 19, 2019
%     getHRmap returns correct map when removeTilt and removeCurvature are applied.
%
% September 21, 2019
%     if resolution is set to be 0, use the current default value
%
% February 8, 2019
%     function mirrorLossN and lossSummary added
%
% January 16, 2019
%        function loss = mirrorLoss( obj, inPortID, absLoss )
%        new argument, absLoss. If true, return value is actual power loss
%
% November 28, 2018
%   In prepare2x2Like, scaling by tilt is modified. This is still on going...
%
% May 18, 2018
%   doRemoveTiltCurvature is made better
%
% April 2, 2018
%   Spherical is the default mirror surface - r^2/2R
%   isSpherical = -1 for simple spherical without Hello-Vinet formula
%
% May 16, 2017
%   thicknessInSubs( n, correctOffcentering )
%   correctOffcentering removed. 
%   Distance between BS and AR side optics (ITMX or SRM3) are the length of optical path.
% 
% April 25, 2017
%   getHRRoC returns thetaX and thetaY
%
% March 1st, 2017
%    printInfo includes ABCD matrixes
%    port ID is printed by name (March 2nd)
%
% September 16, 2016
%    The mirror maps on the AR side were not properly handled.
%    This caused problems for the Virgo ITMs.
%
% July 8, 2016
%    In mirrorLoss, when the port is not connected to anything, don't calculate the source power
%
% June 30, 2016
%    mirrorLoss calculates the difference between (transmitted power + reflected power)
%    and the power going toward the mirror (for the loss of ETM, this is the power from ITM).
%
% September 2, 2015
%    Using new OpticObj structure with ifo property, mostly
%    changing ConstantTK.kLam0 to obj.ifo.fields.kCR
%
% July 8th, 2015
%    delSphPar added to handle spehrical mirror surface more accurately
%       parabalic case : r^2 / 2 RoC
%       spherical case : r^2 / 2 RoC x 2 / ( 1 + sqrt(1 - rsq / RoC^2) )
%
% November 26, 2014
%    applyDisplacement when applied to reflection map itself, the magnitude is compromised.
%    It is applied to the phase and mask.
%
% November 15, 2014
%   applyDisplacement is to displace object by (optX0, optY0).
%   buildMirFFTmaps is modified so that mirrors are displaced.
%   For convenience, phase is calculated which take care of phases of the field.
%   This is moved in the opposite direction to be combined with the original mirror.
%   optTheta and optPhi are implemenetd in buildMirFFTmaps for tilt
%
% October 15, 2014
%   trans() is modified so that, if transMatch(inPortIndex) == inPortIndex, trans acts as refl for back scattering
%
% August 5th, 2014
%   removeCurvature option is added to Mirror2x2
%   doRemoveTilt is replaced by doRemoveTiltCurvature and curvature mismatch is corrected as well
%
% October 22, 2013
% September 12, 2013
%
%
% used in Mirror4x4Obj
%   prepare2x2Like;
%   prepare4x4like;  defined in Mirror4x4Obj
%   buildDefault;
%   buildMirFFTmaps( [1, 2, 3, 3], [1,2;3,4;1,3;2,4] );
%   doRemoveTiltCurvature( true );
%

classdef MirrorObj < OpticObj
    
    properties (SetObservable)
        nRef = ConstantTK.nSubsSiO2;
    end
    
    properties (Constant)
        inputNames0x0 = { 'invRoC', 'Aperture', 'Thickness', 'virtualAperture', ...
            'T', 'Lhr', 'Lar', 'optTheta', 'optPhi', 'optX0', 'optY0', ...
            'isSpherical', 'invFocal', 'resolution', 'WfftUser', 'nRef' };
    end
    
    methods (Static)
        function tagList = getInputNames
            tagList = MirrorObj.inputNames0x0;
        end
    end
    
    properties (Dependent)
        cosAngInSub
        angInSub
        beamShift
    end
    
    properties
        % incident angle, default is normal incident
        incAngle = 0;
        
        % substrate
        Thickness = 0;
        
        % tilt on HR side
        % for now, only 2x2 mirror
        tiltX = 0;
        tiltY = 0;
        
        % displacement / orientation
        %   mech.[x,y,z,tX,tY,tZ]
        
        % optical quantities
        t = 0;
        rHR = -1;
        Rhr0 = 1; % This is used for the round trip calculation. Pure reflectance without loss
        rAR = 1;
        
        % tile definition using eulter angle
        optTheta = 0;   % rotation around y axis after phi roration
        optPhi = 0;     % rotation around z axis
        optDCurv = 0;   % curvature correction to nullify curvature mismatch
        removeTilt = false;  % user flag if alighnment correction is to be done
        removeCurvature = false; % user flag if curvature mismatch correction is to be dobe
        
        invFocal = 0; % focal length (lens = -r^2/2*invF)
        
        % These will be used in mirror2x2 and in mirror4x4
        % mirror4x4 can be specified either by this HR surface (MC end mirror or RC folding mirrors)
        % or by 4 maps, 2 transmission and 2 reflection
        
        % HR surface specification
        HRfiles     % by data file
        % transmisson specification
        TRfiles     % by data file
        % AR surface specification
        % This is used mainly for BS spec to generate mocked up maps
        % For now, it is not used for Mirror2x2
        ARfiles     % by data file
        
        % each are two diemtional, (1) is for input to HR side, (2) is for input to AR side
        reflFFT
        transFFT
        
        % these are used fot debugging
        % to save memory, these can be temporal variable
        RPYmap
        RSXmap
        TPXmap
        TSYmap
        
        % parabolic or spherical mirrors, -1 for simple spherical without Hello-Vinet off-axis correction
        isSpherical = 1;
        
        % refractiove index of HeNe which is used used for map measuremts by Zygo
        nRefMeasure = 1.457
        
        % after all maps are built, this function is called
        % refl are trans are cell(1,4) for mirror4x4 and cell(1,2) for mirror2x2,
        % and each are N/2xN/w maps to be applied to convert in(i) -> out
        % [refl, trans] = modMap( refl, trans, x, y )
        % store handle of this function if necessary
        modMap
        modMapVar = {} % this is passed as the last argument, use it like setting variables
        
        % variables defined when setMisc is called
        % it should be 2N cells, 'name1', val1, 'name2', var2, ...
        % This variables and variables in dataFileManager are not redundant
        % When only name1 is set later, name2=val2 in dataFileManager is used to complete the settings.
        variables
    end
    
    methods
        %% constructor
        function obj = MirrorObj( ifo, nIn, nOut, name, spec )
            
            obj = obj@OpticObj( ifo, name, nIn, nOut);
            
            %            obj.isMirrorObj = true;
            
            spec = obj.defineRTL( spec );
            
            [A, spec] = UtilTK.extractVal( 'Aperture', spec );
            if ~isempty(A)
                obj.ApertureDef = A;
            end
            
            % if resolution is 0, use default value, or do nothing
            defResolution = obj.resolution;
            UtilTK.updateProperties( obj, obj.inputNames0x0, spec );
            if obj.resolution == 0
                obj.resolution = defResolution;
            end
            
            % updateNdep calculates ABCD matrixes
            addlistener(obj, 'nRef', 'PostSet', @obj.updateNdep);
            
            obj.HRfiles = dataFileManager('HRfiles');
            obj.TRfiles = dataFileManager('TRfiles');
            obj.ARfiles = dataFileManager('ARfiles');
            
            obj.variables = cell(0);
        end
        
        % dat(inPort index,1) = port ID
        % dat(inPort index,2) = mirrorLoss of this port
        % dat(end,1) = optics ID
        % dat(end,2) = total loss
        function dat = lossSummary( obj )
            dat = nan(5,2);
            for id = 1 : length( obj.inPorts )
                dat(id,1) = obj.inPorts(id);
                dat(id,2) = obj.mirrorLoss( obj.inPorts(id), true );
            end
            dat(end,1) = obj.ifo.optics.findNamedOptic( obj.name );
            dat(end,2) = sum( dat(~isnan( dat(1:end-1,2) ), 2) );
        end
        
        % calculate r and t from specification
        function spec = defineRTL( obj, spec )
            % T is power transmittance of HR surface
            [T, spec] = UtilTK.extractVal( 'T', spec, 0 );
            [Lhr, spec] = UtilTK.extractVal( 'Lhr', spec, 0 );
            [Lar, spec] = UtilTK.extractVal( 'Lar', spec, 0 );
            
            obj.rHR = 1 - T - Lhr;
            if obj.rHR < 0
                error('T + Loss on HR > 1');
            end
            % 1-Lar is transmittance of AR surface
            % (1-Lar)*Rhr = (1-Lar)*(1-T-Lhr) ~= 1 - Lar - T
            obj.rAR = (1-Lar)*obj.rHR;
            if obj.rAR < 0
                error('T + Loss on AR > 1');
            end
            % T is treated like transmittance of HR surface
            obj.t = T * (1 - Lar);
            if obj.t < 0
                obj.t = 0;
            end
            obj.t = sqrt(obj.t);
            obj.rHR = -sqrt(obj.rHR);
            obj.Rhr0 = 1 - T; % no loss component
            obj.rAR = sqrt(obj.rAR);
        end
        
        %% access functions
        
        % map = getHRmap( obj, removeRoC )
        % Returns RPYmap/2, the reflection map on the HR side
        function map = getHRmap( obj, removeRoC )
            if ~exist( 'removeRoC', 'var' )
                removeRoC = true;
            end
            map = obj.RPYmap/2;
                        
            [x,y,r] = obj.getXYR;
            
            Cthick = cos( obj.optTheta ) - 1;
            Cx = cos( obj.optPhi )*sin( obj.optTheta );
            Cy = sin( obj.optPhi )*sin( obj.optTheta );
            delZ = Cthick * obj.Thickness/2 - Cx * x - Cy * y;
            
            map = map + delZ + r.^2/2*obj.optDCurv;

            if removeRoC
                map = map - (x.^2 + y.^2)*obj.invRoC/2;
            end
        end
        
        % [RoC, piston] = getHRRoC( obj, radius )
        % calculate RoC, thetaX, thetaY and piston using zernikeCoef
        function [RoC, thetaX, thetaY, piston] = getHRRoC( obj, radius )
            [xl,yl] = obj.getXYR;
            [~, coef] = zernikeCoef( obj.RPYmap/2, xl, yl, radius*2 );
            % coef:(1) piston, (2) theta x, (3) theta y, (4) power
            RoC = radius^2/(4*coef(4));
            thetaX = coef(2) / radius;
            thetaY = coef(3) / radius;
            piston = coef(1);
        end
        
        % loss = mirrorLoss4( obj )
        %   loss(:).inPorts(1:2), loss(:).outPorts(1:2) : 
        %        this loss is inPorts(1)+inPorts(2) -> outPort(1)+outPort(2)
        %        For mirror2x2, there is one loss, for mirror4x4, there are two losses
        function loss = mirrorLossN( obj )
            if isa( obj, 'Mirror4x4Obj' )
                loss(1).inPorts = [obj.inPorts(1), obj.inPorts(4)];
                loss(1).outPorts = [obj.outPorts(2), obj.outPorts(3)];
                loss(2).inPorts = [obj.inPorts(2), obj.inPorts(3)];
                loss(2).outPorts = [obj.outPorts(1), obj.outPorts(4)];
            else
                loss(1).inPorts = [obj.inPorts(1), obj.inPorts(2)];
                loss(1).outPorts = [obj.outPorts(1), obj.outPorts(2)];
            end
            for iLoss = 1 : length(loss)
                loss(iLoss).loss = ...
                    obj.ifo.fields.calcPower(loss(iLoss).inPorts(1)) + ...
                    obj.ifo.fields.calcPower(loss(iLoss).inPorts(2)) - ...
                    obj.ifo.fields.calcPower(loss(iLoss).outPorts(1)) - ...
                    obj.ifo.fields.calcPower(loss(iLoss).outPorts(2));
            end
        end
        
        %   loss(:).loss = power(inPort(1)) + power(inPort(2)) - power(outPort(1)) - power(outPort(2))
        % loss = mirrorLoss( obj, inPortID, absLoss )
        %   loss = 1 - ( reflected power + transmitted power ) / input power
        %   if absLoss is true, loss = input power - ( reflected + transmitted ) power
        function loss = mirrorLoss( obj, inPortID, absLoss )
            loss = nan;
            for ip = 1 : length( obj.inPorts )
                if obj.inPorts(ip) == inPortID
                    obj.ifo.fields.calcIfNecessary( inPortID );
                    %   inPower = obj.ifo.fields.fieldList(inPortID).calcSumIn(obj.Aperture/2);
                    srcID = obj.ifo.props.getPropFrom( inPortID );
                    if srcID <= 0  % port not connected, so no source
                        return;
                    end
                    inPower = obj.ifo.fields.fieldList(srcID).calcPower;
                    if inPower == 0
                        return;
                    end
                    outPower = 0;
                    % port to which the field is reflected
                    if length( obj.reflMatch ) >= ip
                        reflPortID = obj.outPorts( obj.reflMatch(ip) );
                        if reflPortID > 0
                            EFFTsave = obj.ifo.fields.fieldList(reflPortID).EFFT;
                            obj.refl( ip, obj.ifo.fields.fieldList(inPortID).EFFT );
                            outPower = obj.ifo.fields.fieldList(reflPortID).calcPower;
                            obj.ifo.fields.fieldList(reflPortID).EFFT = EFFTsave;
                        end
                    end
                    % port to which the field is transmitted
                    if length( obj.transMatch ) >= ip
                        transPortID = obj.outPorts( obj.transMatch(ip) );
                        if transPortID > 0
                            EFFTsave = obj.ifo.fields.fieldList(transPortID).EFFT;
                            obj.trans( ip, obj.ifo.fields.fieldList(inPortID).EFFT );
                            outPower = outPower + obj.ifo.fields.fieldList(transPortID).calcPower;
                            obj.ifo.fields.fieldList(transPortID).EFFT = EFFTsave;
                        end
                    end
                    if exist('absLoss', 'var') && absLoss
                        loss = inPower - outPower;
                    else
                        loss = 1 - outPower / inPower;
                    end
                    return;
                end
            end
        end
        
        
        %% geometrical properties
        
        % angle in the substrate
        function inAng = get.angInSub( obj )
            inAng = asin( sin(obj.incAngle) / obj.nRef );
        end
        
        % cos angle in the substrate
        function cosAng = get.cosAngInSub( obj )
            cosAng = sqrt ( 1 - (sin(obj.incAngle) / obj.nRef)^2 );
        end
        
        % offset of the beam in x of points entering to one side and leaving on the other
        function d = get.beamShift( obj )
            d = obj.Thickness * tan( acos( obj.cosAngInSub ) );
        end
        
        % physical length to go through a substrate
        % n = 1, 0, -1 : length * nRef^n
        %
        % Distance between BS and AR side optics (ITMX or SRM3) are the length of optical path.
        % argument correctOffcentering removed
        function d = thicknessInSubs( obj, n  )
            % d is the geometrical length of the ray in the substrate
            d = obj.Thickness / obj.cosAngInSub;
            if exist( 'n', 'var' )
                switch n
                    case -1
                        d = d / obj.nRef;
                    case 1
                        d = d * obj.nRef;
                end
            end
        end
        
        % clear apertures of four maps
        %  kind       for BS [1,2,3,3],
        %      for mirror2x2 [0,0,0,0]
        function Amask = calcMask( obj, kind )
            if obj.ApertureDef < 0 || kind == 0
                % normal incident case (kind=0) or mirror size is very large (aperture<0)
                Amask = obj.ifo.ffts.zeroMat;
                Amask( obj.getRsq <= (obj.Aperture^2/4) ) = 1;
            else
                % LIGO-P980004 (B. Bachner's thesis) p50
                % oval of one BS surface viewed by one beam
                % some centered, some offset, and one full path is a product of two ovales
                oval = @(xoff) ( ( (obj.ifo.ffts.Xm * obj.Wfft - cos(obj.incAngle)*xoff ) / cos(obj.incAngle) ).^2 + ...
                    (obj.ifo.ffts.Yn * obj.Wfft).^2 ) <= (obj.Aperture^2/4);
                switch kind
                    case 1
                        % BS ref from PRM to Y arm
                        Amask = oval( 0 );
                    case 2
                        % BS ref from SRM to X arm
                        Amask = oval( obj.beamShift ) & oval( -obj.beamShift );
                    case 3
                        % BS trans from PRM to X arm or SRC to Y arm, +x is clipped
                        Amask = oval( -obj.beamShift ) & oval( 0 );
                    case 4
                        % BS trans from Y arm to SRM, -x is clipped
                        Amask = oval( obj.beamShift ) & oval( 0 );
                end
            end
            Amask = im2double( Amask );
        end
        
        %% 
        % function [ RPY, RSX, TPX, TSY, specified ] = prepare2x2Like( obj )
        %
        % this creates maps based on three 2D maps, HR surface, transmission and
        % AR surface maps.
        % The transmission map with finite incident angle is built assuming
        % the substrate non-uniformity is deapth independent, and may not be accurate
        % To handle realistic mirrors with internal structure,
        % specify 2 reflection and 2 transmission maps
        %
        % if no files are specified, it is assumed to be perfect
        
        function [ RPY, RSX, TPX, TSY, specified ] = prepare2x2Like( obj )
            
            specified = 0;
            
            [hrmap, invRoCHR] = obj.HRfiles.loadMaps( obj );
            if isempty( hrmap )
                % if no file is specified
                hrmap = zeros(size(obj.getXYR));
                invRoCHR = obj.invRoC;
            else
                % if any file is specified and yet, RoC is not defined, use the original invRoC
                specified = specified + 1;
                if invRoCHR == 0
                    invRoCHR = obj.invRoC;
                end
            end
            % based on the curvature and incident angle,
            % calculate the surface figure and position dependent reflection angle
            % HR surface = RPYmap + s, s ~ -r^2/RoC
            [s, cosAlpha] = obj.evalAberration( invRoCHR );
            RPY = 2 * hrmap .* cosAlpha - s*cos(obj.incAngle);
            % hrmap has the curvature included
            %            hrmap = hrmap + obj.getRsq/2*invRoCHR*cos(obj.incAngle);
            hrmap = hrmap + MirrorObj.delSphPar( obj.getRsq, invRoCHR, obj.isSpherical )*cos(obj.incAngle);
            
            % transmission map and AR side reflection maps
            % SPTMAP has curvature of HR side subtracted, or HR side curvature is the reference
            % This is true for BS (normal incident) and ITM
            [sptmap, invRoCTransMap]  = obj.TRfiles.loadMaps( obj );
            if isempty( sptmap )
                sptmap = obj.getRsq/2*obj.invFocal;
                %                sptmap = zeros(size(obj.getXYR));
            else
                specified = specified + 2;
                sptmap = sptmap + obj.getRsq/2*invRoCTransMap;
            end
            
            [armap, invRoCARMap] = obj.ARfiles.loadMaps( obj );
            if isempty( armap )
                armap = obj.getRsq/2*obj.invRoCAR;
                %                armap = zeros(size(obj.getXYR));
            else
                specified = specified + 4;
                % AR side reflection is not important, so no s
                % in case there is AR side curvature...
                %                armap = armap + obj.getRsq/2*invRoCARMap;
                armap = armap + MirrorObj.delSphPar( obj.getRsq, invRoCARMap, obj.isSpherical );
            end
            
            % compose the 4 maps for BS out of three 2D maps,
            % HR and AR reflection + transmission maps
            
            if obj.incAngle ~= 0
                % compose additional three maps using hrmap, tmap and armap
                [RSX, TPX, TSY] = obj.tiltMaps( hrmap, sptmap, armap );
                
                % all maps defined wrt original z axis
                % fields at (-2,-1, 0, 1, 2) coming at cos(incident angle)
                % hits the mirror at(-2.-1,0,1,2)/cos(thata) in the original axis
                % or value at N in the tilted axis is the value at N/cos in the normal
                % or value at N*cos for the reflect field is the value at N on the mirror map
                [x0, y0] = obj.getXYR;
%                x1 = x0*cos(obj.incAngle);
                x1 = x0/cos(obj.incAngle);
                % PRY(N) is the value at x1(N*cos) when PRY is given at x0(N)
                RPY = interp2( x0, y0, RPY, x1, y0 ); RPY(isnan(RPY)) = 0;
                RSX = interp2( x0, y0, RSX, x1, y0 ); RSX(isnan(RPY)) = 0;
                TPX = interp2( x0, y0, TPX, x1, y0 ); TPX(isnan(RPY)) = 0;
                TSY = interp2( x0, y0, TSY, x1, y0 ); TSY(isnan(RPY)) = 0;
             %{
                x1 = x0 / cos(obj.incAngle);
                RPY = interp2( x1, y0, RPY, x0, y0 );
                RSX = interp2( x1, y0, RSX, x0, y0 );
                TPX = interp2( x1, y0, TPX, x0, y0 );
                TSY = interp2( x1, y0, TSY, x0, y0 );  
             %}
            else
                % RPY = 2 * hrmap .* cosAlpha - s*cos(obj.incAngle); calculated when hrmap is provided
                % this conversion is done properly in tiltMaps
                sptmap = sptmap * obj.nRef/ obj.nRefMeasure;
                RSX = 2 * ( sptmap - hrmap * obj.nRef - armap * (obj.nRef-1) );
                TPX = sptmap - (obj.nRef-1)*(hrmap + armap);
                TSY = TPX;
            end
        end
        
        %% [RPY, RSX, TPX, TSY ] = buildDefault( obj )
        % build maps when nothing is spcified using parameters in addMirror
        % this is for normal incident case, or very close to it
        % Eout = exp( i k [R,T] ) Ein = exp( - i k delL )
        % [R,T] = -optical path length
        function [RPY, RSX, TPX, TSY ] = buildDefault( obj )
            
            % HR side reflection
            hrmap = obj.getRsq/2*obj.invRoC;
            s = obj.evalAberration( obj.invRoC );  % -r^2/(2 R) * Hello-Vinet correction
%            s = -2*hrmap;  % HY to be restored
            
            % transmission map and AR side reflection maps
            % invFocal > 0 => converging lens
            sptmap = obj.getRsq/2*obj.invFocal;
            
            % AR side spherical map
            armap = obj.getRsq/2*obj.invRoCAR;
            
            % +r^2/2 R = -(-r^2/2 R)
            RPY = -s*cos(obj.incAngle);
            % -(dL of map)
            RSX = 2 * ( sptmap - hrmap*obj.nRef - armap*(obj.nRef-1) );
            TPX = sptmap - (obj.nRef-1)*(hrmap + armap);
            TSY = TPX;
        end
        
        %%
        % Compose BS maps out of 2D transmission maps.
        % The transmission map with finite incident angle is built assuming
        % that the substrate non-uniformity is deapth independent.
        % Clear apertures for a finite size BS need to be multiplied to get correct maps,
        % I.e., maps are as large as data files can be applied.
        %
        % all quantities are calculated using optical path length.
        % At the end of the function, signs are flipped to conform to the Zygo measurement convension
        %
        function [RSX, TPX, TSY] = ...
                tiltMaps( obj, hrmap, sptmap, armap )
            
            % definition of the orientation is different between hrmap and (sptmap, armap)
            % unify to hrmap. So all are facing HR out
            sptmap = fliplr(sptmap);
            armap = fliplr(armap);
            
            % shift of the beam location from oneside to another, in units of the FFT pixel
            nOffset = obj.beamShift / obj.getDXDY;
            
            % substrate optical thickness per unit physical length
            % assmuned the optical thickness is proportinally scaled by the ratio of
            % the nominal refractive index from HeNe to NdYaG
            % data / d0 (value in unit optical path length ) * D (optical thickness of each slice)
            % = data / d0 * d0 / cos(angInSub) / nOffset = data / cos / nOffset
            %
            % sptmap is genative of optical path length
            mapSub = ...
                ( -sptmap - (obj.nRefMeasure-1)*armap - obj.nRefMeasure*hrmap ) ...
                * obj.nRef / obj.nRefMeasure / obj.cosAngInSub / nOffset;
            
            % transmission maps from HR side to AR, TPX toward X Arm, TSY toward SRC
            TPX = zeros(size(mapSub));
            TSY = zeros(size(mapSub));
            
            % sum up layers with optical length D = d0 / cos(angInSub) / nOffset
            for in = 1 : ceil( nOffset )
                if in == ceil(nOffset)
                    weight = nOffset - floor(nOffset);
                else
                    weight = 1;
                end
                TPX(:,in:end) = TPX(:,in:end) + mapSub(:,1:end-in+1) * weight;
                TSY(:,1:end-in+1) = TSY(:,1:end-in+1) + mapSub(:,in:end) * weight;
            end
            
            [xl,yl] = obj.getXYR;
            
            % for TPX case, beam at the center on HR goes to -beamShift (facing to out from HR) on AR
            % i.e., TPX(:,center) uses ARMAP(:,center-beamShift)
            if max(max(abs(armap))) == 0
                armapShift = zeros( size( TPX ) );
            else
                armapShift = interp2( xl, yl, armap, xl-obj.beamShift, yl );
            end
            TPX = TPX + (obj.nRef/obj.cosAngInSub - 1/cos(obj.incAngle)) * armapShift ...
                + obj.nRef/obj.cosAngInSub * hrmap;
            
            % for TSY case, beam at the center on HR goes to +beamShift on AR
            if max(max(abs(armap))) == 0
                armapShift = zeros( size( TPX ) );
            else
                armapShift = interp2( xl, yl, armap, xl+obj.beamShift, yl );
            end
            TSY = TSY + (obj.nRef/obj.cosAngInSub - 1/cos(obj.incAngle)) * armapShift ...
                + obj.nRef/obj.cosAngInSub * hrmap;
            
            % I got clazy about the sign
            RSX = TPX + TSY;
            
            TPX = TPX - hrmap/cos(obj.incAngle);
            TSY = TSY - hrmap/cos(obj.incAngle);
            
            % TPX is a map measued from PRC, which is opposite to all other cases
            TPX = fliplr(TPX);
            
            % flip the sign to change the convension from optical path length to Zygo measurement convension
            RSX = -RSX;
            TPX = -TPX;
            TSY = -TSY;
        end
        
        %% build reflFFT and transFFT from RPY, RSX, TPX and TSY maps
        %
        % for BS (for TM, it is trivial)
        % 1) HR side reflection
        %    in : PRC -> out : Y arm and in : Y arm -> out : PRC
        % 2) AR side reflection
        %    in : X arm -> out SRC and in : SRC -> out X arm
        % 3) transmission between PRC and X arm (HR to AR)
        %    For PRC to X, +x is clipped
        % 4) transmission from SRC to Y arm (AR to HR)
        %    For SRC to y, +x is clipped, so mask 3, same as PRC to x
        
        % maskID : Amask is caluclated by obj.calcMask( masiID(1:4) ),
        %         for BS [1,2,3,3],
        %         for mirror [0,0,0,0]
        % sides : for BS [1,2;3,4;1,3;2,4]
        %         for mirror [1;2;1;2]
        function buildMirFFTmaps( obj, masks, sides )
            
            [xMir, yMir, r2Mir] = obj.getXYR;
            r2Mir = r2Mir.^2;
            if (~obj.removeTilt) && (obj.optTheta ~= 0)
                % rotate around z by phi, then rotate around y by theta
                Cx = cos( obj.optPhi )*sin( obj.optTheta );
                Cy = sin( obj.optPhi )*sin( obj.optTheta );
                HRTiltCor = -2 * (Cx * xMir + Cy * yMir);
            else
                HRTiltCor = 0;
            end
            
            % HR side reflection
            % in PRC -> out Y arm, in Y arm -> out PRC
            Amask = obj.calcMask(masks(1));
            obj.RPYmap = obj.RPYmap .* Amask;
            for iside = sides(1,:)
                inID = obj.inPorts(iside);
                outID = obj.outPorts( obj.reflMatch(iside) );
                del = obj.ifo.fields.getPhiCorr(outID) - obj.ifo.fields.getPhiCorr(inID);
                % s is the change of reflection length, ~ -1/R, twice of hight x -1
                % RPYmap is height*2*cos
                phase = obj.applyDisplacement( HRTiltCor + obj.RPYmap, obj.reflMatch(iside) ) + del / 2 * r2Mir;
                
                % map is defined seeing from PRC
                % +x affects on Ein(-x) for both direction, so no flip
                obj.reflFFT{iside} = obj.rHR * exp( 1i * obj.ifo.fields.kCR * phase ) .* obj.applyDisplacement( Amask, obj.reflMatch(iside) );
            end
            
            % AR side reflection
            % in X arm -> out SRC, in SRC -> out X arm
            Amask = obj.calcMask(masks(2));
            obj.RSXmap = obj.RSXmap .* Amask;
            for iside = sides(2,:)
                inID = obj.inPorts(iside);
                outID = obj.outPorts( obj.reflMatch(iside) );
                % invRoCHR is the HR side curvature, and map is defined relative to that
                del = obj.ifo.fields.getPhiCorr(outID) - obj.ifo.fields.getPhiCorr(inID);
                % RSXmap is the pathlength from AR side to HR side and back to AR
                % Opposite to RPYmap
                % I got almost crazy about the sign
                phase = obj.applyDisplacement( -fliplr(HRTiltCor) + obj.RSXmap, obj.reflMatch(iside) ) + del / 2 * r2Mir;
                % map is defined seeing from SRC
                % +x affects on Ein(-x) for both direction, so no flip
                obj.reflFFT{iside} = obj.rAR * exp( 1i * obj.ifo.fields.kCR * phase ) .* obj.applyDisplacement( Amask, obj.reflMatch(iside) );
            end
            
            % transmission between PRC and X arm (HR to AR)
            % For PRC to X, +x is clipped
            Amask = obj.calcMask(masks(3));
            obj.TPXmap = obj.TPXmap .* Amask;
            for iside = sides(3,:)
                % iside 1 : PRC -> X arm
                % iside 3 : X arm -> PRC
                inID = obj.inPorts(iside);
                outID = obj.outPorts( obj.transMatch(iside) );
                del = obj.ifo.fields.getPhiCorr(outID) - obj.ifo.fields.getPhiCorr(inID);
                phase = obj.applyDisplacement( obj.TPXmap, obj.transMatch(iside) ) + del / 2 * r2Mir;
                % tmap is defined using axis defined by +z pointing out of HR
                % For PRC to X arm, outgoing field and map have opposite x direction
                obj.transFFT{iside} = obj.t * exp( 1i * obj.ifo.fields.kCR * phase ) .* obj.applyDisplacement( Amask, obj.transMatch(iside) );
                if iside == 3
                    obj.transFFT{iside} = fliplr( obj.transFFT{iside} );
                end
            end
            
            % transmission from SRC to Y arm (AR to HR)
            % For SRC to y, +x is clipped, so mask 3, same as PRC to x
            Amask = obj.calcMask(masks(4));
            obj.TSYmap = obj.TSYmap .* Amask;  % this is for debugging
            for iside = sides(4,:)
                % iside 2 : Y arm -> SRC
                % iside 4 : SRC -> Y arm
                inID = obj.inPorts(iside);
                outID = obj.outPorts( obj.transMatch(iside) );
                del = obj.ifo.fields.getPhiCorr(outID) - obj.ifo.fields.getPhiCorr(inID);
                phase = obj.applyDisplacement( obj.TSYmap, obj.transMatch(iside) ) + del / 2 * r2Mir;
                % hrmap and tmap are defined using axis defined by +z pointing out of HR
                % For Y arm to SRC, outgoing field and map have opposite x direction
                obj.transFFT{iside} = obj.t * exp( 1i * obj.ifo.fields.kCR * phase ) .* obj.applyDisplacement( Amask, obj.transMatch(iside) );
                if iside == 2
                    obj.transFFT{iside} = fliplr( obj.transFFT{iside} );
                end
            end
            
        end
        
        %% displace all reflFFT and transFFT maps by optX0 and optY0
        % This is used to shift the beam power term. That moves the beam opposite direction
        function newmap = applyDisplacement( obj, map, outIndex )
            if obj.optX0 == 0 && obj.optY0 == 0
                newmap = map;
                return;
            end
            
            [x,y] = obj.getXYR;
            if obj.isHRsideIndex( outIndex )
                xSign = 1;
            else
                xSign = -1;
            end
            newmap = interp2( x, y, map, x-obj.optX0*xSign, y-obj.optY0 );
        end
        
        %% distance change on reflection
        % s ~= -(x^2/cos(theta) + y^2*cos(theta))/R
        % cosAlpha ~= cos(incAngle)
        function [s, cosAlpha] = evalAberration( obj, invRoCMap )
            % reflection on the HR side
            [x,y] = obj.getXYR;
            switch obj.isSpherical
                case 1
                    [s, cosAlpha] = MirrorObj.Spherical( invRoCMap, x, y, obj.incAngle );
                case 0  % parabolic ay not correct
                    [s, cosAlpha] = MirrorObj.Parabolic( invRoCMap, x, y, obj.incAngle );
                otherwise % simple surface
                    s = -(x.^2/cos(obj.incAngle) + y.^2*cos(obj.incAngle))*invRoCMap;
                    cosAlpha = ones(size(x));
             end
        end
        
        %% calculate reflection by a mirror, in and out are both internal format
        % outField = reflFFT * fliplr( inField )
        % out( +x ) = reflFFT( +x ) * inField( -x )
        %
        % when tiltX or tiltY is given, the reflected field on AR side is tilted so that it goes back to the original direction.
        % When the mirror is tilted, the promptly reflected fields does not go back to the original direction.
        % This tilt action sets the refleced field on the AR side goes back to the incident direction.
        % This is to support a simply mode matching for a FP with a tilted cavity beam axis, like sweet spot seach
        function outField = refl( obj, inPortIndex, inField )
            if ~exist( 'inField', 'var' )
                inField = obj.ifo.fields.fieldList( obj.inPorts(inPortIndex) ).EFFT;
            end
            
            outField = obj.reflInOut( obj.reflFFT{inPortIndex}, inField );

            % only from AR side reflection
            if ( obj.tiltX ~= 0 || obj.tiltY ~= 0 ) && ( obj.reflMatch(inPortIndex) == 2 )
                % for the reflection on the AR side, correct angle so that the direction is along the incident z axis
                outField = obj.ifo.fields.fieldList(obj.outPorts(obj.reflMatch(inPortIndex))).tiltField( -obj.tiltX*2, obj.tiltY*2, outField );
            end
            
            if nargout == 0
                outID = obj.outPorts(obj.reflMatch(inPortIndex));
                obj.ifo.fields.fieldList(outID).EFFT = outField;
                %                obj.ifo.fields.fieldList(outID).status = -1;
            end
        end
        
        %% calculate transmission through a mirror, in and out are both internal format
        % outField = transFFT * inField
        % out( +x ) = reflFFT( +x ) * inField( +x )
        %
        % if transMatch(inPortIndex) == inPortIndex, this is back scattering and reflection formula is applied
        % This is used for Virgo MC end mirror whose back scatter induces fields which circulates in the reverse direction
        %
        % when tiltX or tiltY is given, the infield for HR->AR and the outField for AR->HR are tilted
        % When the mirror is tilted, the transmitted fields changes directions by the tilted mirror angle.
        % This tilt action sets the transmitted field on the HR to AR goes back to the incident direction, and
        % the transmitted field from AR->HR goes to the proper tilted direction in the cavity.
        % This is to support a simply mode matching for a FP with a tilted cavity beam axis, like sweet spot seach.
        %
       function outField = trans( obj, inPortIndex, inField )
            
            if ~exist( 'inField', 'var' )
                inField = obj.ifo.fields.fieldList( obj.inPorts(inPortIndex) ).EFFT;
            end                
            
            if obj.transMatch(inPortIndex) == inPortIndex
                % this should never happen !!!
                % inport == outport is back reflection
                outField = obj.reflInOut( obj.transFFT{inPortIndex}, inField );
            else
                % normal transmission
                
                % if HR side field needs to be tilted
                % input to HR (inPortIndex = 1) or output to HR (transMatch(inPortIndex) = 1)
                if ( obj.tiltX ~= 0 || obj.tiltY ~= 0 ) && ( inPortIndex == 1 || obj.transMatch(inPortIndex) == 1 )                    
                    if inPortIndex == 1  % in to HR, going out of AR
                        doTilt = -1;
                    else  % in to AR, going out of HR
                        doTilt = 1;
                    end
                else
                    doTilt = 0;
                end
                
                if doTilt == -1  % HR -> AR
                    % to make the transmitted direction to incident AR z axis, change the input field direction to HR
                    inField = obj.ifo.fields.fieldList(obj.inPorts(inPortIndex)).tiltField( -obj.tiltX, obj.tiltY, inField );
                end
                
                outField = obj.transInOut( obj.transFFT{inPortIndex}, inField );
                
                if doTilt == 1  % AR -> HR
                    % to make the transmitted field in the HR surface to match with the cavity axis, tilt the out going field on HR side
                    % without this, input field direction needs to be changed so that the transmitted field goes to a proper tilted cavity axis
                    outField = obj.ifo.fields.fieldList( obj.outPorts(obj.transMatch(inPortIndex)) ).tiltField( obj.tiltX, obj.tiltY, outField );
                end
            end
                
            if nargout == 0
                outID = obj.outPorts(obj.transMatch(inPortIndex));
                obj.ifo.fields.fieldList(outID).EFFT = outField;
            end            
        end
        
        %% about this mirror
        function info = printInfo( obj )
            info = printInfo@OpticObj( obj );
            
            lossDat = obj.mirrorLossN;
            info = sprintf('%s\n',info);
            
            for il = 1 : length(lossDat)
                info = sprintf( '%sloss %g in ''%s''+''%s''->''%s''+''%s''', info, lossDat(il).loss, ...
                    obj.ifo.fields.fieldList(lossDat(il).inPorts(1)).name, obj.ifo.fields.fieldList(lossDat(il).inPorts(2)).name, ...
                    obj.ifo.fields.fieldList(lossDat(il).inPorts(1)).name, obj.ifo.fields.fieldList(lossDat(il).inPorts(2)).name );
            end
            
            loss = 1-obj.rHR^2-obj.t^2;
            if loss < 1e-10
                loss = 0;
            end
            info = sprintf( '%s\n\nR = %g, T = %g, L = %g', info, obj.rHR^2, obj.t^2, loss );
            if abs(obj.rHR) ~= abs(obj.rAR)
                info = sprintf( '%s, R(AR) = %g', info, obj.rAR^2 );
            end
            info = sprintf( '%s\n', info );
            if obj.invRoC == 0
                info = sprintf( '%sRoC(HR) = flat', info);
            else
                info = sprintf( '%sRoc(HR) = %g', info, 1/obj.invRoC );
            end
            if obj.invRoCAR ~= 0
                info = sprintf( '%s Roc(AR) = %g', info, 1/obj.invRoCAR );
            end
            if obj.optX0 == 0 && obj.optY0 == 0
                offsetStr = '';
            else
                offsetStr = sprintf(' offset = (%f, %f),', obj.optX0, obj.optY0);
            end
            info = sprintf( '%s, Aperture = %f, Thickness = %f,%s n = %f, Inc.Angle = %g (degree), theta = %g, phi = %g\n', info, ...
                obj.Aperture, obj.Thickness, offsetStr, obj.nRef, obj.incAngle/pi*180, obj.optTheta/pi*180, obj.optPhi/pi*180 );
            
            info = sprintf('%swOpt = %g, Wfft = %g\n', ...
                info, obj.wOpt, obj.Wfft );
            
            info = sprintf('%s\nABCD matrix\n', info);
            for iport = 1 : length( obj.inPorts )
                info = sprintf( '%s   refl  from ''%s''[%2d] to ''%s''[%2d] = [%10.5g, %10.5g; %10.5g %10.5g]\n', info, ...
                    obj.ifo.fields.getName(obj.inPorts(iport)), obj.inPorts(iport), ...
                    obj.ifo.fields.getName(obj.outPorts(obj.reflMatch(iport))), obj.outPorts(obj.reflMatch(iport)), ...
                    obj.reflABCD{iport}([1,3,2,4]) );
            end
            for iport = 1 : length( obj.inPorts )
                info = sprintf( '%s   trans from ''%s''[%2d] to ''%s''[%2d] = [%10.5g, %10.5g; %10.5g %10.5g]\n', info, ...
                    obj.ifo.fields.getName(obj.inPorts(iport)), obj.inPorts(iport), ...
                    obj.ifo.fields.getName(obj.outPorts(obj.transMatch(iport))), obj.outPorts(obj.transMatch(iport)), ...
                    obj.transABCD{iport}([1,3,2,4]) );
            end
            info = sprintf('%s\n', info);
            
            if obj.HRfiles.numFiles > 0
                info = sprintf( '%s%s\n', info, obj.HRfiles.fileInfo );
            end
            
            if obj.TRfiles.numFiles > 0
                info = sprintf( '%s%s\n', info, obj.TRfiles.fileInfo );
            end
            
            if obj.ARfiles.numFiles > 0
                info = sprintf( '%s%s\n', info, obj.ARfiles.fileInfo );
            end
        end
        
        % misc works
        function setupMisc( obj )
            % if there were any default method of setupMisc defined in the superclass, do it
            setupMisc@OpticObj( obj );
            
            % this is the method of this specific object
            obj.updateMirMaps;
        end
        
        %% function updateMirMaps( obj, varargin )
        function updateMirMaps( obj, varargin )
            %% function updateMirMaps( varargin )
            % varargin redefines variables used in the formula provided in dataFileManaher.addMap('fileName', 'formula')
            % varargin is like 'RoC', 100e3, 'delL', 1e-6
            % Variables in varargin override default values provided at the function call
            if ~UtilTK.validArgList( varargin{:} )
                error('variable definition incorrect');
            end
            
            obj.variables = varargin;
            
            % make default reflection and transmission map, which are zeros
            obj.reflFFT = cell(size(obj.inPorts));
            [obj.reflFFT{:}] = deal( obj.ifo.ffts.zeroMat );
            obj.transFFT = cell(size(obj.inPorts));
            [obj.transFFT{:}] = deal( obj.ifo.ffts.zeroMat );
            
            % build maps
            obj.buildMaps;
            
            % aftercare by calling modMap and removing nan in maps
            obj.afterCare;
        end
        
        %% after all maps are built, let modMap to do the final touch, like adding BS baffle
        % nan's are replaced by 0, so reflFFT and transFFT do not have nans
        % reflFFT and transFFT are 0 outside of clear aperture, so outField = mapFFT * inField
        function afterCare( obj )
            if ~isempty( obj.modMap )
                [x, y] = obj.getXYR;
                [obj.reflFFT, obj.transFFT] = obj.modMap( obj.reflFFT, obj.transFFT, x, y, obj.modMapVar{:} );
            end
            
            % cleanup NaN sorunding the boundary (nan*0 = nan, so aperture mask does not clean)
            for id = 1 : length( obj.reflFFT )
                obj.reflFFT{id}( isnan( obj.reflFFT{id} ) ) = 0;
                obj.transFFT{id}( isnan( obj.transFFT{id} ) ) = 0;
            end
        end
        
        %% remote temporal variables for garbage collection
        % remote 4 maps which are not used for field calculation
        function cleanTemp( obj )
            obj.RPYmap = [];
            obj.RSXmap = [];
            obj.TPXmap = [];
            obj.TSYmap = [];
        end
        
        %% rotate optics to make the normally injected field on HR side be reflected without tilt component
        % also curvature mismatch is calculated by using LG10 in the reflected field and minimize curvature mismatch
        % When applyTilt is false, just the tilt correction is calculated, but the mirror is not moved
        % 7/13/2016 : make this to work for Mirror4x4 as well as Mirror2x2
        function corrected = doRemoveTiltCurvature( obj, applyTilt )
            corrected = false;
            
            if ~( obj.removeTilt || obj.removeCurvature )
                return;
            end
            
            obj.optDCurv = 0;
            obj.optTheta = 0;
            obj.optPhi = 0;
            
            HRinID = obj.inPorts(1);
            HRoutID = obj.outPorts( obj.reflMatch(1) );
            
            % save current fields in the input and output ports, so that they will be restored
            inEFFT = obj.ifo.fields.fieldList(HRinID).EFFT;
            outEFFT = obj.ifo.fields.fieldList(HRoutID).EFFT;
                        
            % prepare idealistic field
            obj.ifo.fields.setHGField( HRinID, 0, 0, 1 );
            w = obj.ifo.fields.fieldList(HRoutID).w;
            
            obj.refl( 1 );
            HG00 = obj.ifo.fields.fieldList(HRoutID).HGCoef( 0, 0 );
            
            if obj.removeTilt
                % direction of outgoing field
                % k w theta cos(phi) = i HG10 = -imag( HG10 )
                % k w theta sin(phi) = i HG01 = -imag( HG01 )
                iHG10 = -imag(obj.ifo.fields.fieldList(HRoutID).HGCoef( 1, 0 )/HG00);
                iHG01 = -imag(obj.ifo.fields.fieldList(HRoutID).HGCoef( 0, 1 )/HG00);
                
                theta = sqrt( iHG10^2 + iHG01^2 ) / (obj.ifo.fields.kCR*w);
                obj.optTheta = theta;
                obj.optPhi = atan2( iHG01, iHG10 ) + pi;
                corrected = true;
            end
            
            if obj.removeCurvature
                iLG10 = imag( obj.ifo.fields.fieldList(HRoutID).LGCoef( 1, 0 )/HG00 );
                % LG10 coef = i k w^2/4 (1/Rref - 1/Rmode)
                % 1/Rmir -> 1/Rmir + (1/Rref - 1/Rmode)/2 = 1/Rmir + optDCurv
                obj.optDCurv = iLG10 * 2/( obj.ifo.fields.kCR * w^2 );
                corrected = true;
            end
            
            if applyTilt
                [x,y,r] = obj.getXYR;
                Cthick = cos( obj.optTheta ) - 1;
                Cx = cos( obj.optPhi )*sin( obj.optTheta );
                Cy = sin( obj.optPhi )*sin( obj.optTheta );
                delZ = Cthick * obj.Thickness/2 - Cx * x - Cy * y;
                HRCorr = exp( 2i * obj.ifo.fields.kCR * (delZ  + r.^2/2*obj.optDCurv) );
                ARCorr = exp( 2i * obj.ifo.fields.kCR * (delZ  + r.^2/2*obj.optDCurv*obj.nRef) );
%                 obj.reflFFT{1} = obj.reflFFT{1} .* HRCorr;
%                 obj.reflFFT{2} = obj.reflFFT{2} .* fliplr( conj(ARCorr) );
                
                
                if length(obj.reflFFT) == 2  % Mirror2x2
                    obj.reflFFT{1} = obj.reflFFT{1} .* HRCorr;
                    obj.reflFFT{2} = obj.reflFFT{2} .* fliplr( conj(ARCorr) );
                else % Mirror4x4
                    obj.reflFFT{1} = obj.reflFFT{1} .* HRCorr;                  % HRx in -> HRy out
                    obj.reflFFT{2} = obj.reflFFT{2} .* fliplr( HRCorr );        % HRy in -> HRx out
                    obj.reflFFT{3} = obj.reflFFT{3} .* fliplr( conj(ARCorr) );  % ARx in -> ARy out
                    obj.reflFFT{4} = obj.reflFFT{4} .* conj(ARCorr);            % ARy in -> ARx out
                end
            end

            % restore original fields
            obj.ifo.fields.fieldList(HRinID).EFFT = inEFFT;
            obj.ifo.fields.fieldList(HRoutID).EFFT = outEFFT;
            
        end
    end
    
    methods (Abstract)
        % abstract function, to be implemented by the derived class
        updateNdep( obj )
    end
    
    methods (Static)
        % parabalic case : r^2 / 2 RoC
        % spherical case : r^2 / 2 RoC x 2 / ( 1 + sqrt(1 - rsq / RoC^2) )
        function del = delSphPar( rsq, invRoC, isSpherical )
            del = rsq / 2 * invRoC;
            if isSpherical == 1  % 
                del = del * 2 ./ ( 1 + sqrt( abs(1 - rsq*invRoC^2) ) );
            end
        end
        
        % distance change on reflection
        % s ~= -(x^2/cos(theta) + y^2*cos(theta))/R
        % cosAlpha ~= cos(incAngle)
        function [s, cosAlpha] = Spherical( invR, x, y, theta )
            g = sqrt( cos(theta)^2 + 2*x*sin(theta)*invR - (x.^2+y.^2)*invR^2);
            f = (2*x*sin(theta) - (x.^2+y.^2)*invR) ./ (cos(theta) + g);
            s = f - (x*sin(2*theta) - f*cos(2*theta)) ./...
                (2*sin(2*theta)*(sin(theta)-x*invR).*g - cos(2*theta)*(1-2*g.^2));
            
            cosAlpha = sqrt( cos(theta).^2 + 2*x* sin(theta)*invR - (x.^2+y.^2)*invR^2 );
        end
        
        % s ~= -(x^2 + y^2)/rho
        % cosAlpha = cos( b/R )
        function [s, cosAlpha] = Parabolic( invR, x, y, theta )
            b = 1/invR * tan(theta);
            a = -2*b*invR;
            beta = 1-(b*invR)^2;
            g = 2./(1 + ((x+b).^2 + y.^2)*invR^2);
            f = -(x.^2 + 2*x*b + y.^2)*invR/2;
            s = f + (a*x - beta*f)./(a*g.*(x+b)*invR + beta*(1-g));
            
            cosAlpha = cos( atan( (x+b)*invR ) );
        end
        
        % reflection by mirror = map * fliplr( inField )
        function outField = reflInOut( reflMap, inField )
            outField = reflMap .* fliplr( inField );
        end
        
        % transmission through mirror = map * inField
        function outField = transInOut( transMap, inField )
            outField = transMap .* inField;
        end
        
    end
    
    methods ( Abstract )
        buildMaps( obj )
    end
end
