function inArea = SqInCircle( x, y, rX, rY )
% function inArea = SqInCircle( x, y, rX, rY ) : weight for calcIn or calcOut
%   sum(sum( F(x,y) .* inArea )) is a better approximaton of 
%      sum(sum( F(x,y) ))*dxdy in (x/rX)^2 + y/rY)^2 <= 1 when r > 0
%      sum(sum( F(x,y) ))*dxdy in (x/rX)^2 + y/rY)^2 >= 1 when rX or rY < 0
%
% October 6, 2019 : calculation for calcOut is better than before
%
% Sep. 26, 2019 : if rX or rY < 0, it returns weights to calculate sumOut
%
% Sep. 24, 2019 : (x0,y0) are xvalus and yvalues, not center, corrected
%

if ~exist('rY', 'var')
    rY = rX;
end

calcOut = rX < 0 || rY < 0;

rX = abs(rX);
rY = abs(rY);

if min( size(x) ) == 1 && min( size(y) ) == 1
    [x,y] = meshgrid( x, y );
end

% if quad symmetry exists, calculate only one quadrant
quadSym = mod(size(x,2),2) == 0;
if quadSym
    quadSym = isempty( find( x(1,1:floor(end/2)) ~= -x(1,end:-1:floor(end/2)+1), 1) ) & ...
        isempty( find( y(1:floor(end/2),1) ~= -y(end:-1:floor(end/2)+1, 1), 1 ) );
end

if quadSym
    [x,y] = meshgrid( x(1,1:end/2), y(1:end/2,1) );
end

dxy = x(1,2)-x(1,1);

% use rX and rY as unit
delX = dxy/rX;
delY = dxy/rY;
delXY = delX * delY;

% effectively, flip all to positive side
x = abs(x);
y = abs(y);

% bottom-left corner - closest to the origine
xmin = (x - dxy / 2) / rX;
ymin = (y - dxy / 2) / rY;

% this is the area
if calcOut
    inArea = zeros( size(x) );
    inArea( xmin.^2 + ymin.^2 > 1 ) = delXY;
else
    inArea = ones( size(x) ) * delXY;
    inArea( xmin.^2 + ymin.^2 > 1 ) = 0;
end

% top-right corner - furthest from the origine
% by using min( x, rMax ), the edge effect is automatically taken care
% when the area of ([xmin, xmax], [ymin, ymax]) in a unit circle is calculated.
xmax = xmin + delX;
ymax = ymin + delY;

% if the furthest point is in the region, all are in the region
% so calculate only when the rectangle region is partially included by
% a circle with radius of 1 (normalized by rX and rY)

% nearest point is in, furthest point is out
% so partially in the oval region
onBoundary = find( (xmax.^2 + ymax.^2 >= 1) & (xmin.^2 + ymin.^2 <= 1) );
    
for id = 1 : length(onBoundary)
    index = onBoundary(id);
    topLeftIn = xmin(index)^2 + ymax(index)^2 <= 1;
    botRightIn = xmax(index)^2 + ymin(index)^2 <= 1;
    
    if topLeftIn && botRightIn %% area[11]
        v1 = ymax(index);
        u0 = sqrt( 1 - v1*v1 );
        u1 = xmax(index);
        v0 = sqrt( 1 - u1*u1 );
%        arc = ( (u1-u0)^2 + (v1-v0)^2 )^1.5 / 12;
        theta = atan2(sqrt((u0-u1)^2+(v0-v1)^2), sqrt((u0+u1)^2+(v0+v1)^2));
        arc = theta - sin(2*theta)/2;
        inArea(index) = delXY - (u1-u0)*(v1-v0)/2 + arc;
    elseif topLeftIn % area[10]
        v1 = ymax(index);
        u0 = sqrt( 1 - v1*v1 );
        v0 = ymin(index);
        u1 = sqrt( 1 - v0*v0 );
%        arc = ( (u1-u0)^2 + (v1-v0)^2 )^1.5 / 12;
        theta = atan2(sqrt((u0-u1)^2+(v0-v1)^2), sqrt((u0+u1)^2+(v0+v1)^2));
        arc = theta - sin(2*theta)/2;
        inArea(index) = ((u0+u1)/2 - xmin(index))*delY + arc;
    elseif botRightIn % area[01]
        u1 = xmax(index);
        v0 = sqrt( 1 - u1*u1 );
        u0 = xmin(index);
        v1 = sqrt( 1 - u0*u0 );
%        arc = ( (u1-u0)^2 + (v1-v0)^2 )^1.5 / 12;
        theta = atan2(sqrt((u0-u1)^2+(v0-v1)^2), sqrt((u0+u1)^2+(v0+v1)^2));
        arc = theta - sin(2*theta)/2;
        inArea(index) = ((v0+v1)/2 - ymin(index))*delX + arc;
    else % area[00]
        u0 = xmin(index);
        v1 = sqrt( 1 - u0*u0 );
        v0 = ymin(index);
        u1 = sqrt( 1 - v0*v0 );
%        arc = ( (u1-u0)^2 + (v1-v0)^2 )^1.5 / 12;
        theta = atan2(sqrt((u0-u1)^2+(v0-v1)^2), sqrt((u0+u1)^2+(v0+v1)^2));
        arc = theta - sin(2*theta)/2;
        inArea(index) = (u1-u0)*(v1-v0)/2 + arc;
    end   

    % (u0,v1) - (u1,v0)
    if calcOut
        inArea(index) = delXY - inArea(index);
    end
end

inArea = inArea * rX * rY;

if quadSym
    halfSize = size(inArea,1);
    inArea(1:halfSize,halfSize+1:halfSize*2) = fliplr(inArea);
    inArea(halfSize+1:halfSize*2,1:halfSize*2) = flipud( inArea(1:halfSize,1:halfSize*2));
end

end
