%
% VirgoMC2 : Mirror4x4 with backscattering
%    transmission of the mirror is used for the back scattering.
%    transFFT{1:2} are used for back reflection
%
%                 MC3
%                 ^ v
%                 | |// /
%                 | // /
%         MC1 >--- // / ---> MC1
%             <-- // / <----
%                // / |
%               // /| |
%            HR  AR | ^
%                   v 
%                  MC3
%
% September 2, 2015
%    Using new OpticObj structure with ifo property, mostly
%    changing ConstantTK.kLam0 to obj.ifo.fields.kCR
%
% November 26, 2014
%    map area with no data is filled by 0, so that the reflection map extends to the boundary of mirror
%
% October 13, 2014
%
% Almost normal incident field on HR side, no field on AR side
% The field going into HRx / HRy will reflect out to HRy / HRx, but they can also reflect back to HRx / HRy
% The second reflection is handled by using transmission part
%


classdef VirgoMC2 < Mirror4x4Obj
    
    properties
        
    end
    
    methods
        function obj = VirgoMC2( name, ifo, spec )
            
            obj = obj@Mirror4x4Obj( name, ifo, spec );
            obj.backScatter = true;
            
            % normal 4x4 mirror
            %   refletion HRx-i(1)->HRy-o(2), HRy-i(2)->HRx-o(1), ARx-i(3)->ARy-o(4), ARy-i(4)->ARx-o(3)
            %   transmisn HRx-i(1)->ARx-o(3), HRy-i(2)->ARy-o(4), ARx-i(3)->HRx-o(1), ARy-i(4)->HRy-o(2)
            % change transmission to deal with back scattering
            %   transmisn HRx-i(1)->HRx-o(1), HRy-i(2)->HRy-o(2), ARx-i(3)->ARx-o(3), ARy-i(4)->ARy-o(4)
            %   the ABCD matrix of backscattering HRx-i(1:2)->HRx-o(1:2) is the same as forward scattering HRx-i(1:2)->HRy-o(2:1)
            obj.transMatch = [1,2,3,4];
            obj.transABCD = obj.reflABCD;
            
        end
        %%%{
        function buildMaps( obj )
            % build map by cutting high frequency in map, not amplitude (exp(i k map).
            % this makes the absolute values of R and T under control, i.e., < 1.
            % HR 1 -> HR 2, HR2 -> HR 1, normal reflection
            % HR 1 -> HR 1 and HR 2 -> HR 2, freq shifted reflection
           
            % load map data with the original resolution
            [hrmap, invRoCHR, xMap, yMap] = obj.HRfiles.loadMaps( obj, true );
            if isempty( hrmap )
                % if no file is specified
                hrmap = zeros(length(yMap), length(xMap));
                invRoCHR = obj.invRoC;
            else
                % if any file is specified and yet, RoC is not defined, use the original invRoC
                if invRoCHR == 0
                    invRoCHR = obj.invRoC;
                end
            end
            
            [x0, y0] = obj.getXYR;
            Common = obj.rHR * obj.calcMask(1) .* exp( 1i * obj.ifo.fields.kCR * obj.getRsq * invRoCHR * cos(obj.incAngle) );
            for iang = -1 : 1
%                mapOut = UtilTK.lowPassMap( hrmap, xMap, yMap, x0, y0, 0.35, 20, obj.incAngle, true );
                mapOut = UtilTK.lowPassMap( hrmap, xMap, yMap, x0, y0, 1, 20, obj.incAngle, false );
                mapOut = mapOut + iang * obj.incAngle * x0;
                mapOut( isnan( mapOut ) ) = 0;
                if iang == 0
                    % refl{1} HRx -> HRy
                    del = obj.ifo.fields.getPhiCorr(obj.outPorts(2)) - obj.ifo.fields.getPhiCorr(obj.inPorts(1));
                    obj.reflFFT{1} = Common .* exp( 1i * obj.ifo.fields.kCR * ( 2*mapOut + del / 2 * obj.getRsq ) );
                    % refl{2} HRy -> HRx
                    del = obj.ifo.fields.getPhiCorr(obj.outPorts(1)) - obj.ifo.fields.getPhiCorr(obj.inPorts(2));
                    obj.reflFFT{2} = Common .* exp( 1i * obj.ifo.fields.kCR * ( 2*mapOut + del / 2 * obj.getRsq ) );
                elseif iang == 1
                    % trans{1} HRx -> HRx
                    del = obj.ifo.fields.getPhiCorr(obj.outPorts(1)) - obj.ifo.fields.getPhiCorr(obj.inPorts(1));
                    obj.transFFT{1} = Common .* exp( 1i * obj.ifo.fields.kCR * ( 2*mapOut + del / 2 * obj.getRsq ) );
                else
                    % trans{1} HRy -> HRy
                    del = obj.ifo.fields.getPhiCorr(obj.outPorts(2)) - obj.ifo.fields.getPhiCorr(obj.inPorts(2));
                    obj.transFFT{2} = Common .* exp( 1i * obj.ifo.fields.kCR * ( 2*mapOut + del / 2 * obj.getRsq ) );
                end
            end
        end
        %%%}
        
        %{
        % this is old style. 
        % If high frequency cutoff is applied in reflection/transmission map itself, 
        % the absolute values of refelection / transmission can become larger than unity.
        function buildMaps( obj )
            
            % load map data with the original resolution
            [hrmap, invRoCHR, xMap, yMap] = obj.HRfiles.loadMaps( obj, true );
            if isempty( hrmap )
                % if no file is specified
                hrmap = zeros(length(yMap), length(xMap));
                invRoCHR = obj.invRoC;
            else
                % if any file is specified and yet, RoC is not defined, use the original invRoC
                if invRoCHR == 0
                    invRoCHR = obj.invRoC;
                end
            end
            
            % make size be multiple of 4
            indx = find( abs(xMap) < obj.Aperture/2*0.999 );
            if mod(length(indx),2) ~= 0
                indx = indx(1:end-1);
            end
            indy = find( abs(yMap) < obj.Aperture/2*0.999 );
            if mod(length(indy),2) ~= 0
                indy = indy(1:end-1);
            end
            lengxy = 4*floor(min( length(indx), length(indy) )/4);
            indx = indx( (length(indx)-lengxy)/2+1 : (length(indx)+lengxy)/2 );
            indy = indy( (length(indy)-lengxy)/2+1 : (length(indy)+lengxy)/2 );
            xMap = xMap(indx);
            yMap = yMap(indy);
            hrmap = hrmap( indy, indx );
            hrmap(isnan(hrmap)) = 0;
            
            [x0, y0] = obj.getXYR;
            hrmap = hrmap * cos(obj.incAngle);
            [xMesh,~] = meshgrid( xMap, yMap );
            Common = obj.rHR * obj.calcMask(1) .* exp( 1i * obj.ifo.fields.kCR * obj.getRsq * invRoCHR * cos(obj.incAngle) );
            leng0 = length( find( x0(1,:)) >= xMap(1) & x0(1,:) <= xMap(end) );

            if leng0 / lengxy < 0.5
                yIn = (xMap(end)-xMap(1))/(2*leng0)*(-leng0+1:2:leng0-1);
                xIn = yIn * cos( obj.incAngle );
                [xIn, yIn] = meshgrid( xIn, yIn );
            else
                [xIn, yIn] = meshgrid( xMap*cos( obj.incAngle ), yMap );
            end
            
            for iang = -1:1
                tmp = exp(2i * obj.ifo.fields.kCR * (hrmap + iang*obj.incAngle*xMesh));
                if leng0 / lengxy < 0.5
                    tmp = fftshift(fft2(ifftshift(tmp)));
                    tmp = tmp(1+(lengxy-leng0)/2:(lengxy+leng0)/2,1+(lengxy-leng0)/2:(lengxy+leng0)/2);
                    tmp = tmp.*UtilTK.lowpassFilter(size(tmp,2), size(tmp,1), 0.35, 20);
                    tmp = fftshift(ifft2(ifftshift(tmp)))*(leng0/lengxy)^2;
                end
                tmp = interp2( xIn, yIn, tmp, x0, y0 );
                tmp = tmp .* Common;
                if iang == 0
                    % refl{1} HRx -> HRy
                    del = obj.ifo.fields.getPhiCorr(obj.outPorts(2)) - obj.ifo.fields.getPhiCorr(obj.inPorts(1));
                    obj.reflFFT{1} = tmp .* exp( 1i * obj.ifo.fields.kCR * del / 2 * obj.getRsq );
                    % refl{2} HRy -> HRx
                    del = obj.ifo.fields.getPhiCorr(obj.outPorts(1)) - obj.ifo.fields.getPhiCorr(obj.inPorts(2));
                    obj.reflFFT{2} = tmp .* exp( 1i * obj.ifo.fields.kCR * del / 2 * obj.getRsq );
                elseif iang == 1
                    % trans{1} HRx -> HRx
                    del = obj.ifo.fields.getPhiCorr(obj.outPorts(1)) - obj.ifo.fields.getPhiCorr(obj.inPorts(1));
                    obj.transFFT{1} = tmp .* exp( 1i * obj.ifo.fields.kCR * del / 2 * obj.getRsq );
                else
                    % trans{1} HRy -> HRy
                    del = obj.ifo.fields.getPhiCorr(obj.outPorts(2)) - obj.ifo.fields.getPhiCorr(obj.inPorts(2));
                    obj.transFFT{2} = tmp .* exp( 1i * obj.ifo.fields.kCR * del / 2 * obj.getRsq );
                end
            end
        end
        %}
    end
end
