% function [data, xl, yl] = loadASCIIGNT(filename)
%
% Load ASCII GINT data files with the following data structure
%
%--------------------------
% 0.27083725 1001 [pixel X spacing, index 1001]
% 0.27912733 1002 [pixel Y spacing, index 1002]
% 444.0000 1003 [number of columns, index 1003]
% 430.0000 1004 [number of rows, index 1004]
% 0.0000000 1005 [X coordinate at map center, index 1005]
% 0.0000000 1006 [Y coordinate at map center, index 1006]
% -4.025665e-005 1007 [Z minimum, index 1007]
% 4.087238e-005 1008 [Z maximum, index 1008]
% 0.0000000 1009
% 0.0000000 1010
% 1.000000 1011 [X-axis unit (mm), index 1011]
% 1.000000 1012 [Y-axis unit (mm), index 1012]
% 1.000000 1013 [Z-axis unit (mm), index 1013]
% 9999.000 1014 [Empty-pixel value, index 1014]
% 0.0000000 1015
% 0.0000000 1016
% 0.0000000 1017
% 9999.000 1 [First pixel (empty), row 1]
% 9999.000 1 [Second pixel (still empty), row 1]
% 9999.000 430 [Made it through all the pixels]
% -1.000000 -1 [Redundant end-of-data line]
% more lines, but neglected
%--------------------------
%
% Hiro Yamamoto
% September 3rd, 2015
%
function [data, xl, yl] = loadASCIIGNT(filename)

% open the file
[fid, message]=fopen(filename,'r');
if fid == -1
    error('File (''%s'') open failed because %s', filename, message);
end

% read the data
[data,count] = fscanf(fid,'%e %d ');
if (count < 14) || (mod(count,2) ~= 0)
    error('data file did not have proper header information');
end

% close the file
fclose(fid);

% data(2,:) is index
% data(1,:) is data
data = reshape( data, 2, count/2);

% map(Ny, Nx)
Nx = data(1, 3);
Ny = data(1, 4);

% center of the optic in the map
% If (0,0), they need to be adjusted
x0 = data(1, 5);
y0 = data(1, 6);

% data with this value are not measured, so it means NaN
dataNaN = data(1, 14);

% bin size, same for x and y
XbinSize = data(1, 1) * data(1, 11) * 0.001;
YbinSize = data(1, 2) * data(1, 12) * 0.001;

% create the data array
xl = ((1:Nx) - Nx/2 - 0.5) * XbinSize - x0;
yl = ((1:Ny) - Ny/2 - 0.5) * YbinSize - y0;

% data range
iBegin = find( data(2,:) == 1, 1 );
iEnd = find( data(2,:) == -1, 1, 'last' ) - 1;

if (iEnd - iBegin + 1) < Nx*Ny
    error('The number of data stored (%d) is less than expected (%d)', count, Nx*Ny );
end

zScale = data(1, 13) * 0.001;
data = data(1,iBegin:iEnd);
data( data == dataNaN ) = NaN;
data = data * zScale;

% data(1:Nx*Ny) = data[ (x,y)=(1,1), (2,1), (3,1), (1,2), (2,2), (3,2) ]
% data = reshape( Nx, Ny ) => 
%    11    12
%    21    22
%    31    32
% data = data'
%    11    21    31
%    12    22    32
% flipup(data)
%    12    22    32
%    11    21    31

data = flipud(reshape(data, Nx, Ny)');
