%
% function [cwR, info, Efit] = modeFitData( xl, yl, data, k )
%
% input : 1D array (1xN) of x value list
%         2D complex field (NxN), standard format
%
% output : fit parameters of the data by gaussian form
%     data = gauss1D(x) * gauss1D(y)
%     gauss1D(x) = exp( -(x-c0)^2 ( 1/w^2 + i invRk / 2 )
%
% March 1, 2019
%     The range of data used for gauss fit is changed to use data with amplitude > (max + min)/2.
%     When there is a tiny structure at the center, the fit used only several data.
%
% July 30, 2018
%     gaussFit string returns z value
%
% May 18, 2018
%     Minor modification
%
% September 18, 2016
%     fitSummary (wx,wy) shows (abs(wx),abs(wy)) with  i prepended if w is imaginary
%     If the center part is local minimal, w becomes imaginary.
%     Example : (wx,wy) = (0.0476161, i0.122838)
%
% April 21, 2016
%     cwR(1) and cwR(2) swapped. data(jy,ix) contains E(y(jy),x(ix))
%     With the original definition, cwR(1/2) had informaiton about y/x.
%     Now, cwR(1/2) has information about x/y.
%
% September 2, 2015
%   Argument k added
% 
% September 19, 2014
%   info text is added in the return value, which used to be in FieldObj.gaussFit
%
% March 24th, 2014 HY
%   2nd return value added, which is the gaussian field using the fit parameter
%

function [cwR, info, Efit] = modeFitData( xl, yl, EFFT, k )

    cwR = repmat( struct('c0', 0, 'w', 0, 'invRk', 0), 1, 2 );

    if isempty( EFFT )
        return;
    end

    N = length( xl );
    N2 = N/2;

    % average amplitude at the center of the field
    E0amp = sum( sum( EFFT( N2:N2+1, N2:N2+1 ) ) ) / 4;

    % go for x and y
    for ixy = 1 : 2

        fldPower = [0,0];
        fldPhase = [0, 0];
        lastPhase = [0, 0];
        delPhase = [0, 0];

        if ixy == 1
            dx = xl(2)-xl(1);
            dat = [(EFFT(N2,N2+1:end) + EFFT(N2+1,N2+1:end)); ...
                (EFFT(N2,N2:-1:1) + EFFT(N2+1,N2:-1:1)) ] / (2 * E0amp);
        else
            dx = yl(2)-yl(1);
            dat = [(EFFT(N2+1:end,N2) + EFFT(N2+1:end,N2+1)).' ; ...
                (EFFT(N2:-1:1,N2) + EFFT(N2:-1:1,N2+1)).'] / (2 * E0amp);
        end

        % use data in the central region with enough power and the phase sign does not change
        numData = 0;
        Pwr2 = 0; Pwr1 = 0; Pwr0 = 0;
        Pha2 = 0; Pha1 = 0; Pha0 = 0;

        % the range of data used for fit is defined by the magnitude of the data
        for iside = 1 : 2
            halfVal = ( min(abs(dat(iside,1:N2))) + max(abs(dat(iside,1:N2))) ) / 2;
            if abs(dat(iside,1)) > halfVal
                nmax(iside) = max( find( abs(dat(iside,1:N2)) < halfVal, 1 ), 3 );
            else
                nmax(iside) = max( find( abs(dat(iside,1:N2)) < halfVal, 1 ), 3 );
            end
        end
        
        for in2 = 1:N2
            for iside = 1:2
                goodStatus = dat(iside,in2) ~= 0;
                if ~goodStatus
                    break;
                end
                fldPower(iside) = 2 * log( norm( dat(iside,in2) ) );
                goodStatus = fldPower(iside) >= log(0.01);
                if ~goodStatus
                    break;
                end

                lastPhase(iside) = fldPhase(iside);
                fldPhase(iside) = myunwrap( dat(iside,in2), lastPhase(iside) );

                goodStatus = true;
                % avoid the first case
                % central points might be flat, r^2, so do only 4th data and after
                if lastPhase(iside) ~= 0 && in2 > nmax(iside)
                    % slope is meaningful only after 2 data handled
                    if delPhase(iside) ~= 0
                        % if the slope is different, stop it
                        goodStatus = (fldPhase(iside) - lastPhase(iside))*delPhase(iside) >= 0;
                        if ~goodStatus
                            break;
                        end
                    end
                    delPhase(iside) = fldPhase(iside) - lastPhase(iside);
                end
            end
            if ~goodStatus
                break;
            end

            numData = numData + 1;
            xn = dx * (in2 - 0.5);
            Pwr2 = Pwr2 + xn^2 * (fldPower(1) + fldPower(2));
            Pwr1 = Pwr1 + xn * (fldPower(1) - fldPower(2));
            Pwr0 = Pwr0 + fldPower(1) + fldPower(2);
            Pha2 = Pha2 + xn^2*(fldPhase(1) + fldPhase(2));
            Pha1 = Pha1 + xn*(fldPhase(1) - fldPhase(2));
            Pha0 = Pha0 + fldPhase(1) + fldPhase(2);
        end

        X2 = dx^2 * numData*(4*numData^2-1)/6;
        X4 = X2*dx^2*(12*numData^2-7)/20;
        X0 = 2*numData;

        A00 = X0/(X0*X4 - X2^2);
        A02 = X2/(X2^2 - X0*X4);
        A11 = (X0*X4 - X2^2)/(X2*(X0*X4-X2^2));

        C2 = A00*Pwr2 + A02*Pwr0;
        C1 = A11 * Pwr1;
        C3 = A00 * Pha2 + A02 * Pha0;
        cwR(ixy).c0 = -C1/(2*C2);
        cwR(ixy).w = sqrt(-2/C2);
        cwR(ixy).invRk = -2*C3;
    end

    if nargout >= 2
        if ~exist( 'k', 'var' )
            k = ConstantTK.kLam0;
        end
        info = fitSummary( cwR, k );
        w = (cwR(1).w + cwR(2).w)/2;
        invR = (cwR(1).invRk + cwR(2).invRk)/2/k;
        invQ = invR - 2i/(k*w^2);
        info = sprintf( '%s, z = %g, Power = %g', info, real(1/invQ), UtilTK.sumIn(xl, EFFT) );
    end
    
    if nargout == 3
        % calculate the field
        [XX, YY] = meshgrid( xl, yl );
        Efit = E0amp * ...
            exp( -(XX-cwR(1).c0).^2 * ( 1/cwR(1).w^2 + 1i * cwR(1).invRk / 2 ) ) .* ...
            exp( -(YY-cwR(2).c0).^2 * ( 1/cwR(2).w^2 + 1i * cwR(2).invRk / 2 ) );
    end
end

function info = fitSummary( cwR, k )
eps = 1e-6;
notSame =@ ( a, b ) abs(a-b) > eps*(abs(a) + abs(b));

R = 1./[cwR(:).invRk] * k;
if notSame( cwR(1).w, cwR(2).w ) || notSame( R(1), R(2) )
    if real(cwR(1).w) == 0
        cwRXIma = 'i';
    else
        cwRXIma = '';
    end
    if real(cwR(2).w) == 0
        cwRYIma = 'i';
    else
        cwRYIma = '';
    end
    info = sprintf( '(x,y) = (%g, %g), (wx,wy) = (%s%-13.10g, %s%-13.10g), (Rx, Ry) = (%-13.10g, %-13.10g)', ...
        cwR(1).c0, cwR(2).c0, cwRXIma, abs(cwR(1).w), cwRYIma, abs(cwR(2).w), R(1), R(2) );
else
    info = sprintf( '(x,y) = (%g, %g), w = %-13.10g, R = %-13.10g', ...
        cwR(1).c0, cwR(2).c0, (cwR(1).w + cwR(2).w)/2, (R(1) + R(2))/2 );
end
end

function phase = myunwrap( amp, refPhase )
newPhase = atan2( imag(amp), real(amp) );
Nwrap = round( refPhase/(2*pi) );
delRefPhase = refPhase - 2*pi*Nwrap;
delNewPhase = newPhase - 2*pi*round(newPhase/(2*pi));
diff = delNewPhase - delRefPhase;
if diff < -pi
    Nwrap = Nwrap + 1;
elseif diff >= pi
    Nwrap = Nwrap - 1;
end

phase = delNewPhase + 2*pi*Nwrap;
end
